"""This module provides a simpler way to create GUIs for command-line programs.

Overview
--------

Command-line programs typically accept some number of *options*, many of which
must be followed by some number of *arguments*. The arguments commonly represent
filenames, numbers, text, fonts or colors, all of which should be displayed in
a particular way in a GUI interface. This module was designed with many common
option and argument types in mind.

Each command-line option, and any arguments expected by it, are encapsulated in
a `~control.Control`. Specialized controls exist for various kinds of options;
for instance, one of the most common kinds is a `~control.Flag`, which simply
enables or disables certain behavior. For instance, the ``ls`` program has an
``-l`` flag for enabling a "long" output format. In metagui, this might be
declared like::

    Flag('Long listing', '-l')



It assumes your GUI is a direct frontend to one or more command-line programs,
with each command-line option having an associated GUI control widget. Several
kinds of widget are provided, for setting Filename, Color, Number or Font, or
for picking a Choice or Flag.

You probably know a handful of command-line applications that could benefit
from a GUI, even a cheesy-looking Tkinter one. This module makes it easy to
create one.


Define controls
---------------

Say, if you have a program that takes input and output filenames::

    $ tovid -in movie.avi -out movie_encoded

then you can create GUI widgets for those options like this::

    Filename('Input filename', '-in')
    Filename('Output prefix', '-out')

These have the general format::

    Control('Label', '-option', ...)

where:

    ``Control``
        is a ``Control`` subclass appropriate to whatever is being controlled
    ``'Label'``
        is the text that should appear next to the GUI Control
    ``'-option'``
        is a command-line option whose value is set by the Control

Other parameters include default value, help/tooltip text to show, allowable
values, and hints about how to draw the GUI control widget; they are specific
to the flavor of Control being used.

For a full list of available Control subclasses and how to use them,
see the `~libtovid.metagui.control` module.


Creating the GUI
----------------

Controls can be grouped together either vertically or horizontally (using a
`~libtovid.metagui.panel.VPanel` or `~libtovid.metagui.panel.HPanel`, respectively)::

    general = VPanel('General',
        Filename('Background audio file', '-bgaudio'),
        Flag('Create submenus', '-submenus'),
        Number('Length of menu (seconds)', '-menu-length', 0, 120)
        )

This will create three GUI widgets in a panel labeled 'General': one for typing
or browsing to a filename, one for enabling or disabling submenus, and another
for setting menu length to a number between 0 and 120. You can nest panels
inside one another for grouping; sub-panels have their own label and list of
Controls or sub-Panels.

Once you have a panel, you can create an `~libtovid.metagui.gui.Application`::

    app = Application('todisc', [general])

This says your application will run the 'todisc' command-line program, passing
options set by the 'General' panel. Now, create the
`~libtovid.metagui.gui.GUI`::

    gui = GUI('MyGUI', [app])
    gui.run()

This creates the GUI, draws all the widgets, and will run your command-line
program at the push of a button.


Create a multi-panel GUI
------------------------

If your program has a lot of options, one panel may not be enough to hold them
all without looking cluttered, so you may break them down into multiple Panels,
which will be shown in the GUI as tabs that you can switch between. Create
Panels like this::

    thumbs = Panel('Thumbnails',
        Color('Mist color', '-thumb-mist-color', ...),
        Text('Wave', '-wave', ...)
    )
    text = Panel('Text',
        Font('Menu font face', '-menu-font', ...),
        Number('Menu font size', '-menu-fontsize', ...)
    )

then, create the Application and GUI::

    todisc = Application('todisc', [thumbs, text])
    gui = GUI('MyGUI', [todisc])
    gui.run()

If multiple panels are given to Application, a tabbed interface is created,
with one tab for each panel.

"""

# Export everything from support, control, panel, and gui modules
# (I'm open to suggestion about more elegant ways to do this...)
from libtovid.metagui.support import *
from libtovid.metagui.control import *
from libtovid.metagui.panel import *
from libtovid.metagui.gui import *

from libtovid.metagui.support import __all__ as all_support
from libtovid.metagui.control import __all__ as all_control
from libtovid.metagui.panel import __all__ as all_panel
from libtovid.metagui.gui import __all__ as all_gui
__all__ = [
    'variable',
    'support',
    'control',
    'gui',
    'log',
    'manpage',
    'builder',
    'tooltip',
    'widget'] + all_support + all_control + all_panel + all_gui




"""This module does colored console output by calling color-named functions.
To use it, simply::

    print(green("Looking good"))
    print(red("Uh-oh..."))

It was stolen from an early version of a Gentoo module called ``output.py``,
copyright 1998-2003 Daniel Robbins, Gentoo Technologies, Inc., distributed
under the GNU GPL v2::

    # $Header: /home/cvsroot/gentoo-src/portage/pym/output.py,v 1.16 \
    2003/05/29 08:34:55 carpaski Exp $

Modified for inclusion in libtovid.
"""

__all__ = [
    'color',
    'ctext',
    'bold',
    'teal',
    'turquoise',
    'pink',
    'purple',
    'blue',
    'darkblue',
    'green',
    'darkgreen',
    'yellow',
    'brown',
    'red',
    'darkred',
]

codes = {
    'reset':     '\x1b[0m',
    'bold':      '\x1b[01m',
    'teal':      '\x1b[36;06m',
    'turquoise': '\x1b[36;01m',
    'pink':      '\x1b[35;01m',
    'purple':    '\x1b[35;06m',
    'blue':      '\x1b[34;01m',
    'darkblue':  '\x1b[34;06m',
    'green':     '\x1b[32;01m',
    'darkgreen': '\x1b[32;06m',
    'yellow':    '\x1b[33;01m',
    'brown':     '\x1b[33;06m',
    'red':       '\x1b[31;01m',
    'darkred':   '\x1b[31;06m',
}

_do_color = True

def color(do_color):
    """Turn colored output on or off at a global level.
    
        do_color
            ``True`` to enable colored output, ``False`` to disable

    """
    global _do_color
    _do_color = do_color

def ctext(color, text):
    """Return a string containing text in the given color.
    """
    if _do_color:
        return codes[color] + text + codes['reset']
    else:
        return text

def bold(text):
    return ctext('bold', text)

def teal(text):
    return ctext('teal', text)

def turquoise(text):
    return ctext('turquoise', text)

def pink(text):
    return ctext('pink', text)

def purple(text):
    return ctext('purple', text)

def blue(text):
    return ctext('blue', text)

def darkblue(text):
    return ctext('darkblue', text)

def green(text):
    return ctext('green', text)

def darkgreen(text):
    return ctext('darkgreen', text)

def yellow(text):
    return ctext('yellow', text)

def brown(text):
    return ctext('brown', text)

def red(text):
    return ctext('red', text)

def darkred(text):
    return ctext('darkred', text)



# __init__.py (guis)

"""This module contains metagui-based GUIs for several of the tovid scripts.

Each of these is intended to be imported and run from some command-line
frontend script (such as ``todiscgui``), but the .py files in this directory
are themselves executable.

Having all GUI components encapsulated in a module makes it easier for
GUIs to borrow parts from each other (such as ``todisc.py`` borrowing
Controls from ``todisc.py``).

Each ``.py`` file in this directory (except this one) should define a
``run()`` method, which starts up the corresponding GUI::

    def run():
        app = Application('tovid', MAIN, VIDEO, AUDIO, BEHAVIOR)
        gui = GUI("tovid metagui", 640, 720, app)
        gui.run()

To make the ``.py`` itself executable, just add::

    if __name__ == '__main__':
        run()


"""

__all__ = [
    'todisc',
    'tovid',
    'makemenu',
    'makexml',
    'idvid',
    #'wxgtk', # for old tovid GUI, currently in libtovid.gui?
]



# makexml.py

from libtovid.metagui import *

_dvd = Flag("DVD", '-dvd', True)
_vcd = Flag("VCD", '-vcd', False)
_svcd = Flag("SVCD", '-svcd', False)

_overwrite = Flag("Overwrite", '-overwrite', False,
    "Overwrite any existing .xml output file")
_quiet = Flag("Quiet", '-quiet', False,
    "Limit output to essential messages")

# TODO: Figure out how to handle multiple cases
# (video list, -menu <menu> video list, -slides <image list> etc.)
# and how to deal with -group / -endgroup
VIDEOS = List("Video files", '', None,
    "List of .mpg video files to include on the disc",
    Filename())

_menu = Filename("Menu file", '-menu', '',
    "Menu .mpg, linking to one or more videos.")

_slides = List("Slideshow images", '-slides', '',
    "Create a slide-show of still image files.",
    Filename())

_group = ''
_endgroup = ''

_topmenu = Filename("Top menu", '-topmenu', '',
    "Menu .mpg to use for the top (root-level) DVD menu. This menu "
    "should link to each of the other menus.")

_titlesets = Flag("Create titlesets", '-titlesets', False,
    "(DVD only) Forces the creation of a separate titleset per title. "
    "This is useful if the titles of a DVD have different video formats, "
    "e.g. PAL + NTSC or 4:3 + 16:9. If used with menus, there must be a "
    "-topmenu option that specifies a menu file with an entry for each "
    "of the titlesets.")

_chapters = Number("Chapter interval", '-chapters', 5,
    "(DVD only) Creates a chapter every INTERVAL minutes. This option "
    "can be put at any position in a <file list> and is valid for "
    "all subsequent titles until a new -chapters option is encountered. "
    "Using this option may take some time, since the duration of the "
    "video must be calculated.", 1, 60, 'minutes')

_nochapters = Flag("No chapters", '-nochapters', False,
    "Don't add any chapter points to the video.")

OUT_PREFIX = Filename("Output file", '', '',
    "Prefix of .xml file to write as output", 'save',
    "Choose an output prefix")

def run():
    pass

if __name__ == '__main__':
    run()




"""A generic GUI widget class, wrapper for tk.Frame.
"""

__all__ = ['Widget']

import time

# Python < 3.x
try:
    import Tkinter as tk
# Python 3.x
except ImportError:
    import tkinter as tk

# Python 3 compatibility
from libtovid import basestring

class Widget (tk.Frame):
    """Generic metagui widget, suitable for Controls, Panels, etc.
    """
    def __init__(self, name='', **kwargs):
        """Create a Widget.

            name
                Unique name for the widget, or '' for an anonymous widget
        """
        if not isinstance(name, basestring):
            raise TypeError("Widget name must be a string.")

        self.name = name
        self.enabled = True
        self.is_drawn = False
        self.kwargs = kwargs


    def draw(self, master, **kwargs):
        """Initialize the base tk.Frame class.
        """
        tk.Frame.__init__(self, master)
        self.is_drawn = True


    def destroy(self):
        """Destroy the widget.
        """
        tk.Frame.destroy(self)
        self.is_drawn = False


    def get_args(self):
        """Return a list of command-line options for this widget.
        """
        return []


    def set_args(self, args):
        """Set widget options from the given list of command-line arguments,
        and remove any successfully parsed options and arguments from ``args``.
        Override this in derived classes.
        """
        pass


    def enable(self, enabled=True):
        """Enable or disable the Widget and all its children.
        """
        self.enabled = enabled
        # Enable/disable all child widgets that allow state changes
        for widget in self.winfo_children():
            if 'state' in widget.config():
                if enabled:
                    widget.config(state='normal')
                else:
                    widget.config(state='disabled')


    def disable(self):
        """Disable the Widget.
        """
        self.enable(False)


    def blink(self):
        """Cause the Widget to "blink" by briefly changing its background color.
        """
        bgcolor = self.cget('background')
        self.config(background='#C0C0F0')
        self.update()
        time.sleep(1)
        self.config(background=bgcolor)




"""Top-level GUI classes.
"""

__all__ = [
    'Executor',
    'Application',
    'GUI',
]

import os
import tempfile
import shlex
import subprocess
from sys import exit, argv

# Python < 3.x
try:
    import Tix
    import Tkinter as tk
    from ScrolledText import ScrolledText
    from tkMessageBox import showinfo, showerror
    from tkFileDialog import (
      asksaveasfilename, askopenfilename
      )

# Python 3.x
except ImportError:
    import tkinter as tk
    import tkinter.tix as Tix
    from tkinter.scrolledtext import ScrolledText
    from tkinter.messagebox import showinfo, showerror
    from tkinter.filedialog import (
      asksaveasfilename, askopenfilename
      )

from libtovid import cli
from libtovid.metagui.widget import Widget
from libtovid.metagui.panel import Panel, Tabs
from libtovid.metagui.control import Control, NoSuchControl
from libtovid.metagui.support import (
  ConfigWindow, Style, ensure_type, askyesno, get_photo_image, show_icons
  )


DEFAULT_CONFIG = os.path.expanduser('~/.metagui/config')

class Executor (Widget):
    """Executes a command-line program, shows its output, and allows input.
    """
    def __init__(self, name='Executor'):
        Widget.__init__(self, name)
        self.command = None
        self.outfile = None
        # Defined in draw()
        self.text = None
        self.callback = None
        self.stdin_text = None
        self.kill_button = None
        self.save_button = None


    def draw(self, master):
        """Draw the Executor in the given master widget.
        """
        Widget.draw(self, master)
        # Log output text area
        self.text = ScrolledText(self, width=1, height=1,
            highlightbackground='gray', highlightcolor='gray', relief='groove',
            font=('Nimbus Sans L', 12, 'normal'))
        self.text.pack(fill='both', expand=True)
        # Bottom frame to hold the next four widgets
        self.frame = tk.Frame(self)
        # Text area for stdin input to the program
        label = tk.Label(self.frame, text="Input:")
        label.pack(side='left')
        self.stdin_text = tk.Entry(self.frame)
        self.stdin_text.pack(side='left', fill='x', expand=True)
        self.stdin_text.bind('<Return>', self.send_stdin)
        # Button to stop the process
        self.kill_button = tk.Button(self.frame, text="Kill", command=self.kill)
        self.kill_button.pack(side='left')
        # Button to save log output
        self.save_button = tk.Button(self.frame, text="Save log",
                                     command=self.save_log)
        self.save_button.pack(side='left')
        # Pack the bottom frame
        self.frame.pack(anchor='nw', fill='x')
        # Disable stdin box and kill button until execution starts
        self.stdin_text.config(state='disabled')
        self.kill_button.config(state='disabled')


    def send_stdin(self, event):
        """Send text to the running program's stdin.
        """
        text = self.stdin_text.get()
        # Write the entered text to the command's stdin pipe
        #self.command.proc.stdin.write(text + '\n')
        # python 3 compatibility
        self.command.proc.stdin.write(bytearray(text + "\n", "ascii"))
        self.command.proc.stdin.flush()
        # Show what was typed in the log window
        self.write(text)
        # Clear the stdin box
        self.stdin_text.delete(0, 'end')


    def execute(self, command, callback=None):
        """Execute the given `~libtovid.cli.Command`, and call the given callback when done.
        """
        if not isinstance(command, cli.Command):
            raise TypeError("execute() requires a Command instance.")
        if not callable(callback):
            raise TypeError("execute() callback must be callable")

        self.command = command
        self.callback = callback

        # Temporary file to hold stdout/stderr from command
        name = self.command.program
        self.outfile = tempfile.NamedTemporaryFile(
            mode='a+', prefix=name + '_output')

        # Run the command, directing stdout/err to the temporary file
        # (and piping stdin so send_stdin will work)
        output = open(self.outfile.name, 'w')
        self.command.run_redir(stdin=subprocess.PIPE,
                               stdout=output,
                               stderr=output)

        # Enable the stdin entry box and kill button
        self.stdin_text.config(state='normal')
        self.kill_button.config(state='normal')

        # Set focus in the stdin entry box
        self.stdin_text.focus_set()

        # Poll until command finishes (or is interrupted)
        self.poll()


    def kill(self):
        """Kill the currently-running command process.
        """
        if self.command:
            self.notify("Killing command: %s" % self.command)
            self.command.kill()
        self.outfile.close()


    def poll(self):
        """Poll for process completion, and update the output window.
        """
        # Read from output file and print to log window
        try:
            data = self.outfile.read()
            if data:
                # Split on newlines (but not on \r)
                lines = data.split('\n')
                for line in lines:
                    self.write(line)
        except ValueError:
            pass

        # Stop if command is done, or poll again
        if self.command.done():
            self.notify("Done executing!")
            self.kill_button.config(state='disabled')
            self.stdin_text.config(state='disabled')
            self.callback()
            self.outfile.close()
        else:
            self.after(100, self.poll)


    def notify(self, text):
        """Print a notification message to the Executor.
        Adds newlines and brackets.
        """
        self.write("\n[[ %s ]]\n" % text)


    def write(self, line):
        """Write a line of text to the end of the log.
        If the line contains ``\\r``, overwrite the current line.
        """
        if '\r' in line:
            curline = self.text.index('end-1c linestart')
            for part in line.split('\r'):
                if part.strip(): # Don't overwrite with an empty line
                    self.text.delete(curline, curline + ' lineend')
                    self.text.insert(curline, part.strip())
        else:
            self.text.insert('end', '\n' + line)

        self.text.see('end')


    def clear(self):
        """Clear the log window.
        """
        self.text.delete('1.0', 'end')


    def save_log(self):
        """Save log window contents to a file.
        """
        filename = asksaveasfilename(parent=self,
            title='Save log window output',
            initialfile='%s_output.log' % self.name)
        if filename:
            outfile = open(filename, 'w')
            outfile.write(self.text.get('1.0', 'end'))
            outfile.close()
            self.notify("Output saved to '%s'" % filename)


from textwrap import dedent
class Application (Widget):
    """Graphical frontend for a command-line program
    """
    def __init__(self, program, *panels):
        """Define a GUI application frontend for a command-line program.

            program
                Command-line program that the GUI is a frontend for
            panels
                One or more Panels, containing controls for the given
                program's options

        After defining the Application, call run() to show/execute it.
        """
        # Ensure that one or more Panel instances were provided
        ensure_type("Application needs Panel instances", Panel, *panels)

        # Initialize
        Widget.__init__(self)
        self.program = program
        self.showing = False
        self.frame = None
        self.panels = list(panels)
        # Add a LogViewer as the last panel
        self.executor = Executor("Log")
        self.panels.append(self.executor)
        # Defined in draw()
        self.tabs = None
        # Defined in draw_toolbar
        self.toolbar = None


    def draw(self, master):
        """Draw the Application in the given master.
        """
        Widget.draw(self, master)
        self.master = master
        # Draw all panels as tabs
        self.tabs = Tabs('', *self.panels)
        self.tabs.draw(self)
        self.tabs.pack(anchor='n', fill='both', expand=True)
        # need to wait until draw is called before setting this variable
        self.script = tk.StringVar()
        self.script.set('')


    def draw_toolbar(self, config_function, exit_function):
        """Draw a toolbar at the bottom of the application.
        """
        self.toolbar = Widget()
        self.toolbar.draw(self)
        # Create the buttons
        config_button = tk.Button(
            self.toolbar, text="Config", command=config_function)
        run_button = tk.Button(
            self.toolbar, text="Run %s now" % self.program, command=self.execute)
        save_button = tk.Button(
            self.toolbar, text="Save script", command=self.prompt_save_script)
        load_button = tk.Button(
            self.toolbar, text="Load script", command=self.prompt_load_script)
        exit_button = tk.Button(
            self.toolbar, text="Exit", command=exit_function)
        # Pack the buttons
        config_button.pack(anchor='w', side='left', fill='x')
        load_button.pack(anchor='w', side='left', fill='x')
        run_button.pack(anchor='w', side='left', fill='x', expand=True)
        save_button.pack(anchor='w', side='left', fill='x')
        exit_button.pack(anchor='e', side='right', fill='x')
        # Pack the toolbar
        self.toolbar.pack(fill='x', anchor='sw')
        # hack to allow wizard to use less clicks and remove uneeded commmands
        if os.getenv('METAGUI_WIZARD'):
            save_button.pack_forget()
            load_button.pack_forget()
            load_button.pack(anchor='w', side='left', fill='x', expand=True)
            save_button.pack(anchor='w', side='left', fill='x', expand=True)
            save_button.config(command=self.save_exit, text='Save to wizard')
            run_button.pack_forget()
            #exit_button.pack_forget()


    def get_args(self):
        """Get a list of all command-line arguments from all panels.
        """
        args = []
        for panel in self.panels:
            args += panel.get_args()
        return args


    def set_args(self, args):
        """Load application settings from a list of command-line arguments.
        The list of args is modified in-place; if all args are successfully
        parsed and loaded, the list will be empty when this method returns.
        """
        for panel in self.panels:
            panel.set_args(args)


    def execute(self):
        """Run the program with all the supplied options.
        """
        self.toolbar.disable()
        self.executor.clear()

        # Get args and assemble command-line
        args = self.get_args()
        command = cli.Command(self.program, *args)

        # Display the command to be executed
        self.executor.notify("Running command: " + str(command))

        # Show the Executor panel
        self.tabs.activate(self.executor)

        # Show prompt asking whether to continue
        if askyesno(message="Run %s now?" % self.program):
            self.executor.execute(command, self.toolbar.enable)
        else:
            self.executor.notify("Cancelled.")
            self.toolbar.enable()

    def set_scriptname(self, name):
        """Set script filename.  Called externally, this sets the variable
        that used used for save_exit() as well as the 'initial file' for
        prompt_save_script().
        """
        self.script.set(name)


    def prompt_save_script(self):
        """Prompt for a script filename, then save the current
        command to that file.
        """
        # TODO: Make initialfile same as last saved script, if it exists
        filename = asksaveasfilename(parent=self,
            title="Select a filename for the script",
            #initialfile='%s_commands.bash' % self.program)
            initialfile=self.script.get() or '%s_commands.bash' % self.program)

        if not filename:
            return

        try:
            self.save_script(filename)
        except:
            showerror(title="Error", message="Failed to save '%s'" % filename)
            raise
        else:
            saved_script = os.path.split(filename)[1]
            showinfo(title="Script saved", message="Saved '%s'" % filename)


    def save_exit(self):
        """Save current command to script and exit, without prompt
        """
        # hack for titleset wizard to get gui's screen position
        from sys import stderr
        stderr.write("%s %+d%+d" 
          %('gui position', self._root().winfo_x(), self._root().winfo_y()))

        default = os.getcwd() + os.path.sep + '%s_commands.bash' % self.program
        filename = self.script.get() or default
        self.save_script(filename)
        exit()


    def prompt_load_script(self):
        """Prompt for script filename; reload gui with current Control settings
        """
        #"""Prompt for a script filename, then load current Control settings
        #from that file.
        #"""
        # Hack: This is broken (commented section), presumably because of
        # load_args limitations. This hack fixes it for libtovid, but does not
        # belong in metagui.  So refactor this when load_args is fixed.
        # see https://github.com/tovid-suite/tovid/issues/121
        filename = askopenfilename(parent=self,
            title="Select a shell script or text file to load",
            filetypes=[('Shell scripts', '*.sh *.bash'),
            ('All files', '*.*')])

        if not filename:
            return
        # begin Hack
        geometry = self._root().winfo_x(), self._root().winfo_y()
        command = argv[0]
        try:
            cmd = command, '--position', '+%s+%s' %geometry, filename
            os.execlp(cmd[0], *cmd)
        except:
            showerror(title="Error", message="Failed load '%s'" %filename)
            raise
        #try:
        #    self.reset()
        #    self.load_script(filename)
        #except:
        #    showerror(title="Error", message="Failed to load '%s')
        #    raise
        #else:
        #    showinfo(title="Script loaded", message="Loaded '%s'" % filename)


    def save_script(self, filename):
        """Save the current command as a bash script.
        """
        args = self.get_args()
        command = cli.Command(self.program, *args)
        outfile = open(filename, 'w')
        outfile.write(command.to_script())
        outfile.close()


    def reset(self):
        """Reset all controls back to their defaults.
        """
        # Clear all controls of existing values
        for control in Control.all.values():
            control.reset()


    def load_script(self, filename):
        # this is now disabled and its functionality passed to guis.helpers.py
        # see https://github.com/tovid-suite/tovid/issues/121
        pass
        #"""Load current Control settings from a text file.
        #"""
        # Read lines from the file and reassemble the command
        #command = ''
        #for line in open(filename, 'r'):
        #    line = line.strip()
            # Discard comment lines and PATH assignment
        #    if line and not (line.startswith('#') or line.startswith('PATH=')):
        #        command += line.rstrip('\\')
        # Split the full command into arguments, according to shell syntax
        #args = shlex.split(command)

        # Ensure the expected program is being run
        #program = args.pop(0)
        #if program != self.program:
        #    raise ValueError("This script runs '%s', expected '%s'" %
        #          (program, self.program))
        # Load the rest of the arguments
        #self.load_args(args)
        #print("Done reading '%s'" % filename)


    # TODO: Relocate all this stuff into Control and its subclasses
    def load_args(self, args):
        """Load settings from a list of command-line arguments.
        """
        # Examine each arg to find those that are option strings
        while args:
            arg = args.pop(0)
            # See if this is a valid option string; if not,
            # print an error and continue
            try:
                control = Control.by_option(arg)
            except NoSuchControl:
                print("Unrecognized argument: %s" % arg)
                continue
            else:
                print("Found option: %s" % arg)

            # If this control expects a boolean option,
            # it's probably a flag
            if control.vartype == bool:
                print("  Setting flag to True")
                control.set(True)
                control.enabler()

            # If this control expects a list, get all
            # the arguments for the list and set them
            elif control.vartype == list:
                items = []
                while args and not args[0].startswith('-'):
                    items.append(args.pop(0))
                print("  Setting list to: %s" % items)
                control.set(items)

            # Otherwise, set a single-value option
            elif control.vartype in (int, float, str):
                value = args.pop(0)
                print("  Setting value to: %s" % value)
                control.set(control.vartype(value))


# This uses Tix.Tk as a base class, to allow Tix widgets within
# (since for some reason a Tix.ComboBox doesn't like to be inside
# a Tkinter.Tk root window)
class GUI (Tix.Tk):
    """GUI with one or more Applications
    """
    def __init__(self, title, width, height, application, **kwargs):
        """Create a GUI for the given application.

            title
                Text shown in the title bar
            width
                Initial width of GUI window in pixels
            height
                Initial height of GUI window in pixels
            application
                Application to show in the GUI

        Keywords arguments accepted:

            inifile
                Name of an .ini-formatted file with GUI configuration
            icon
                Full path to an image to use as the titlebar icon
            position
                position on screen: '+X+Y'
        """
        Tix.Tk.__init__(self)

        # Ensure that one or more Application instances were provided
        ensure_type("GUI needs Application", Application, application)
        self.application = application

        # Get keyword arguments
        self.inifile = kwargs.get('inifile', DEFAULT_CONFIG)
        self.icon_file = kwargs.get('icon', None)
        self.position = kwargs.get('position', '')

        # Set main window attributes
        self.geometry("%dx%d%s" % (width, height, self.position))
        self.title(title)
        self.width = width
        self.height = height

        self.style = Style()
        if os.path.exists(self.inifile):
            print("Loading style from the config file: '%s'" % self.inifile)
            self.style.load(self.inifile)
        else:
            print("Creating config file: '%s'" % self.inifile)
            self.style.save(self.inifile)

        # Show hidden file option in file dialogs
        self.tk.call('namespace', 'import', '::tk::dialog::file::')
        self.tk.call('set', '::tk::dialog::file::showHiddenBtn',  '1')
        self.tk.call('set', '::tk::dialog::file::showHiddenVar',  '0')

        # Handle user closing window with window manager or ctrl-q
        self.protocol("WM_DELETE_WINDOW", self.confirm_exit)
        self.bind('<Control-q>', self.confirm_exit)


    def run(self, args=None, script=''):
        """Run the GUI and enter the main event handler.
        This function does not return until the GUI is closed.
        args: arguments to the programe being run
        script: a filename to save the GUI command output to
        """
        self.draw()
        
        if args:
            # If first argument looks like a filename, load
            # a script from that file
            if os.path.isfile(args[0]):
                filename = args.pop(0)
                print("Loading script file '%s'" % filename)
                try:
                    self.application.load_script(filename)
                except:
                    print("!!! Failed to load '%s'" % filename)
                    raise
                else:
                    print(":-) Successfully loaded '%s'" % filename)

            # If more args remain, load those too
            if args:
                self.application.load_args(args)
            # if a script/project name was provided, set it in 'application'
        if script:
            self.application.set_scriptname(script)

        # Enter the main event handler
        self.mainloop()


    def draw(self):
        """Draw widgets.
        """
        self.style.apply(self)
        self.frame = tk.Frame(self, width=self.width, height=self.height)
        self.frame.pack(fill='both', expand=True)
        self.resizable(width=True, height=True)
        # Draw the application
        self.application.draw(self.frame)
        self.application.pack(anchor='n', fill='both', expand=True)
        # Draw the toolbar
        self.application.draw_toolbar(self.show_config, self.confirm_exit)
        # Set the application icon, if provided
        if self.icon_file:
            show_icons(self, self.icon_file)


    def show_config(self):
        """Open the GUI configuration dialog.
        """
        config = ConfigWindow(self, self.style)
        if config.result:
            self.style = config.result
            print("Saving configuration to: '%s'" % self.inifile)
            self.style.save(self.inifile)
            showinfo("Configuration saved",
                "Saved configuration to '%s'. Please restart the GUI for the"
                " changes to take effect. You may want to 'Save script' before"
                " quitting." % self.inifile)



    def confirm_exit(self, event=None):
        """Exit the GUI, with confirmation prompt.
        """
        # Hack to sync titleset wizard's position with exiting GUI's position
        if os.getenv('METAGUI_WIZARD'):
            from sys import stderr
            stderr.write("%s %+d%+d" 
            %('gui position', self._root().winfo_x(), self._root().winfo_y()))
        if askyesno(message="Exit?"):
            self.quit()





# tovid_batch.py

from libtovid.metagui import *
from libtovid.guis import tovid

_infiles = List('Input files', '-infiles', '',
    'List of video files to encode',
    Filename())

MAIN = VPanel('Main', _infiles, tovid.BASIC_OPTS)

def run():
    app = Application('tovid-batch',
        MAIN,
        tovid.VIDEO,
        tovid.AUDIO,
        tovid.BEHAVIOR)
    gui = GUI("tovid-batch metagui", 640, 720, app)
    gui.run()

if __name__ == '__main__':
    run()




# Stolen from: http://tkinter.unpythonic.net/wiki/ToolTip
"""
Michael Lange <klappnase at 8ung dot at>

The ToolTip class provides a flexible tooltip widget for Tkinter; it is
based on IDLE's ToolTip module which unfortunately seems to be broken
(at least the version I saw).

INITIALIZATION OPTIONS:

    anchor
        where the text should be positioned inside the widget; must be one
        of "n", "s", "e", "w", "nw" and so on; default is "center"
    bd
        borderwidth of the widget; default is 1 (NOTE: don't
        use "borderwidth" here)
    bg
        background color to use for the widget; default is
        "lightyellow" (NOTE: don't use "background")
    delay
        time in ms that it takes for the widget to appear on the
        screen when the mouse pointer has entered the parent
        widget; default is 1500
    fg
        foreground (i.e. text) color to use; default is "black"
        (NOTE: don't use "foreground")
    follow_mouse
        if set to 1 the tooltip will follow the mouse pointer
        instead of being displayed outside of the parent widget;
        this may be useful if you want to use tooltips for
        large widgets like listboxes or canvases; default is 0
    font
        font to use for the widget; default is system specific
    justify
        how multiple lines of text will be aligned, must be
        "left", "right" or "center"; default is "left"
    padx
        extra space added to the left and right within the widget;
        default is 4
    pady
        extra space above and below the text; default is 2
    relief
        one of "flat", "ridge", "groove", "raised", "sunken" or
        "solid"; default is "solid"
    state
        must be "normal" or "disabled"; if set to "disabled" the
        tooltip will not appear; default is "normal"
    text
        the text that is displayed inside the widget
    textvariable
        if set to an instance of Tkinter.StringVar() the variable's
        value will be used as text for the widget
    width
        width of the widget; the default is 0, which means that
        "wraplength" will be used to limit the widgets width
    wraplength
        limits the number of characters in each line; default is 150

WIDGET METHODS:

    configure(``**opts``)
        change one or more of the widget's options as described
        above; the changes will take effect the next time the
        tooltip shows up; NOTE: follow_mouse cannot be changed
        after widget initialization

Other widget methods that might be useful if you want to subclass ToolTip:

    enter
        callback when the mouse pointer enters the parent widget
    leave
        called when the mouse pointer leaves the parent widget
    motion
        is called when the mouse pointer moves inside the
        parent widget if follow_mouse is set to 1 and the
        tooltip has shown up to continually update the
        coordinates of the tooltip window
    coords
        calculates the screen coordinates of the tooltip window
    create_contents
        creates the contents of the tooltip window (by default a Tkinter.Label)
"""

__all__ = ['ToolTip']

try:
    import Tkinter as tk
except ImportError:
    import tkinter as tk

class ToolTip:
    """A Tooltip widget for Tkinter.
    """
    _defaults = {
        'anchor': 'center',
        'bd': 1,
        'bg': '#ffffe0',
        'fg': '#000000',
        'follow_mouse': 0,
        'font': None,
        'justify': 'left',
        'padx': 4,
        'pady': 2,
        'relief': 'solid',
        'state': 'normal',
        'textvariable': None,
        'width': 0,
        'wraplength': 150
    }

    def __init__(self, master, text='Your text here', delay=1500, **opts):
        self.master = master
        self._opts = self._defaults.copy()
        self._opts.update({'delay': delay, 'text': text})
        self.configure(**opts)
        self._tipwindow = None
        self._id = None
        self._id1 = self.master.bind("<Enter>", self.enter, '+')
        self._id2 = self.master.bind("<Leave>", self.leave, '+')
        self._id3 = self.master.bind("<ButtonPress>", self.leave, '+')
        self._follow_mouse = 0
        if self._opts['follow_mouse']:
            self._id4 = self.master.bind("<Motion>", self.motion, '+')
            self._follow_mouse = 1


    def configure(self, **opts):
        """Update tooltip configuration.
        """
        for key in opts:
            if self._opts.has_key(key):
                self._opts[key] = opts[key]
            else:
                raise KeyError('Unknown option: "%s"' % key)

    ### Event handlers
    def enter(self, event=None):
        """Called when mouse pointer enters the parent widget.
        """
        self._schedule()


    def leave(self, event=None):
        """Called when mouse pointer leaves the parent widget.
        """
        self._unschedule()
        self._hide()


    def motion(self, event=None):
        """Called when mouse pointer moves inside the tooltip region.
        """
        if self._tipwindow and self._follow_mouse:
            x, y = self.coords()
            self._tipwindow.wm_geometry("+%d+%d" % (x, y))

    ### Helpers
    def _schedule(self):
        """Schedule the tooltip for display after ``delay`` milliseconds.
        """
        self._unschedule()
        if self._opts['state'] == 'disabled':
            return
        self._id = self.master.after(self._opts['delay'], self._show)


    def _unschedule(self):
        id = self._id
        self._id = None
        if id:
            self.master.after_cancel(id)


    def _show(self):
        """Show the tooltip.
        """
        if self._opts['state'] == 'disabled':
            self._unschedule()
            return
        if not self._tipwindow:
            self._tipwindow = tw = tk.Toplevel(self.master)
            # hide the window until we know the geometry
            tw.withdraw()
            tw.wm_overrideredirect(1)

            if tw.tk.call("tk", "windowingsystem") == 'aqua':
                tw.tk.call("::tk::unsupported::MacWindowStyle",
                           "style", tw._w, "help", "none")

            self.create_contents()
            tw.update_idletasks()
            x, y = self.coords()
            tw.wm_geometry("+%d+%d" % (x, y))
            tw.deiconify()


    def _hide(self):
        """Hide the tooltip.
        """
        tw = self._tipwindow
        self._tipwindow = None
        if tw:
            tw.destroy()


    ##these methods might be overridden in derived classes:##

    def coords(self):
        """Return the (x, y) coordinates of the tip window.
        """
        # The tip window must be completely outside the master widget;
        # otherwise when the mouse enters the tip window we get
        # a leave event and it disappears, and then we get an enter
        # event and it reappears, and so on forever :-(
        # or we take care that the mouse pointer is always outside the
        # tipwindow :-)
        tw = self._tipwindow
        twx, twy = tw.winfo_reqwidth(), tw.winfo_reqheight()
        w, h = tw.winfo_screenwidth(), tw.winfo_screenheight()
        # calculate the y coordinate:
        if self._follow_mouse:
            y = tw.winfo_pointery() + 20
            # make sure the tipwindow is never outside the screen:
            if y + twy > h:
                y = y - twy - 30
        else:
            y = self.master.winfo_rooty() + self.master.winfo_height() + 3
            if y + twy > h:
                y = self.master.winfo_rooty() - twy - 3
        # we can use the same x coord in both cases:
        x = tw.winfo_pointerx() - twx / 2
        if x < 0:
            x = 0
        elif x + twx > w:
            x = w - twx
        return x, y

    def create_contents(self):
        opts = self._opts.copy()
        for opt in ('delay', 'follow_mouse', 'state'):
            del opts[opt]
        label = tk.Label(self._tipwindow, **opts)
        label.pack()





"""Supporting classes for metagui
"""

__all__ = [
    # Functions
    'exit_with_traceback',
    'ensure_type',
    'divide_list',
    'get_photo_image',
    # Others
    'ScrollList',
    'DragList',
    'ComboBox',
    'FontChooser',
    'PopupScale',
    'ConfigWindow',
    'ScrolledWindow',
    'Style',
]

import os
import sys
import traceback
import math
import base64
import re

# Python < 3.x
try:
    import Tkinter as tk
    from tkSimpleDialog import Dialog
    import tkMessageBox
    from ConfigParser import ConfigParser
    import Tix
    import tkFont

# Python 3.x
except ImportError:
    import tkinter as tk
    from tkinter.simpledialog import Dialog
    import tkinter.messagebox as tkMessageBox
    from configparser import ConfigParser
    import tkinter.tix as Tix
    import tkinter.font as tkFont

# Python 2.4
try:
    all
except NameError:
    def all(iterable):
        for element in iterable:
            if not element:
                return False
        return True

from libtovid import cli
from libtovid.util import imagemagick_fonts
from libtovid.metagui.variable import ListVar

### --------------------------------------------------------------------
### Supporting functions
### --------------------------------------------------------------------
def exit_with_traceback(error_message):
    """Exit with a traceback, and print a custom error message.
    """
    # Print helpful info from stack trace
    last_error = traceback.extract_stack()[0]
    traceback.print_stack()
    filename, lineno, foo, code = last_error

    print("Error on line %(lineno)s of %(filename)s:" % vars())
    print("    %s" % code)
    print(error_message)
    sys.exit(1)


def ensure_type(message, required_type, *objects):
    """Ensure that the given objects are of the required type.
    If not, print a message and call `exit_with_traceback`.
    """
    for obj in objects:
        if not isinstance(obj, required_type):
            type_message = "Expected %s, got %s instead" % \
                         (required_type, obj.__class__.__name__)
            exit_with_traceback(type_message + '\n' + message)


def divide_list(items, pieces):
    """Divide a list of things into several pieces, of roughly equal size.
    """
    # Can't divide into fewer than 2 pieces
    if pieces < 2:
        return [items]
    else:
        # N = number of items in each piece
        N = int(math.ceil(float(len(items)) / pieces))
        # For each piece p, extract N items
        return [items[N*p : N*p+N] for p in range(pieces)]


def askyesno(title=None, message=None, **options):
    """Show a popup with Yes / No options. Return ``True`` if Yes was clicked,
    ``False`` if No was clicked.

    This function is to work around a bug in Tkinter 8.5 / Python 2.6, where
    the askyesno function always returns ``False`` after a file dialog has been
    displayed.
    """
    result = tkMessageBox._show(title, message,
        tkMessageBox.QUESTION, tkMessageBox.YESNO,
        **options)
    if str(result) == 'yes':
        return True
    else:
        return False


def get_photo_image(filename, width=0, height=0, background='', dither=False):
    """Get a ``tk.PhotoImage`` from a given image file.

        filename
            Full path to the image file, in any format that 'convert'
            understands
        width
            Desired image width in pixels
        height
            Desired image height in pixels
        background
            An '#RRGGBB' string for the desired background color.
            Only effective for images with transparency.
        dither
            True to dither the image. May increase graininess.
            Smallish images usually look better without dithering.

    Width and height may be used to scale the image in different ways:

        width == 0, height == 0
            Preserve the original image's size
        width > 0, height == 0
            Resize to the given width; height automatically
            adjusts to maintain aspect ratio
        width == 0, height > 0
            Resize to the given height; width automatically
            adjusts to maintain aspect ratio
        width > 0, height > 0
            Resize to exactly the given dimensions
    """
    # Convert the image file to gif using 'convert'
    cmd = cli.Command('convert')

    # Pre-processing
    # Dither if requested, otherwise disable
    if dither:
        cmd.add('-dither')
    else:
        cmd.add('+dither')
    # Set the background color, if provided
    if background:
        cmd.add('-background', background)

    cmd.add(filename)

    # Post-processing
    # Scale appropriately
    if width > 0 and height == 0:
        cmd.add('-resize', '%dx' % width)
    elif width == 0 and height > 0:
        cmd.add('-resize', 'x%d' % height)
    elif width > 0 and height > 0:
        cmd.add('-resize', '%dx%d!' % (width, height))

    # If background color was set, flatten the image
    if background:
        cmd.add('-flatten')

    # Convert to gif data
    cmd.add('gif:-')

    # Run the command, capturing the gif data from standard output
    cmd.run(capture=True, silent=True)
    gif_data = cmd.get_output()
    # Create and return the PhotoImage from the gif data
    return tk.PhotoImage(data=base64.b64encode(gif_data))


def show_icons(window, image):
    '''Show window manager icons for window, if supported.
    The icon argument is full path to the image to be used.
    '''
    icon = get_photo_image(image, 32, 32)
    try:
        window.tk.call('wm', 'iconphoto', window._w, icon)
    except tk.TclError:
        print("Could not set window manager icons.  Maybe your "
              "Tkinter (Tcl) is too old? Continuing anyway.")



### --------------------------------------------------------------------
### Classes
### --------------------------------------------------------------------

class ScrollList (tk.Frame):
    """A Listbox with a scrollbar.

    Similar to a ``tk.Listbox``, a ScrollList shows a list of items.
    ``tk.Variable`` objects may be associated with both the list of items, and
    the one that is currently selected.
    """
    def __init__(self, master=None, items=None,
                 selected=None, **kwargs):
        """Create a ScrollList widget.

            master
                Tkinter widget that will contain the ScrollList
            items
                ListVar or Python list of items to show in listbox
            selected
                Tk StringVar to store currently selected choice in

        Keyword arguments:

            width
                Width of the listbox, in characters
            height
                Height of the listbox, in characters
        """
        tk.Frame.__init__(self, master)

        if type(items) == list:
            items = ListVar(self, items)
        self.items = items or ListVar()

        self.selected = selected or tk.StringVar()
        self.curindex = 0
        self.linked = None
        # Draw listbox and scrollbar
        self.scrollbar = tk.Scrollbar(self, orient='vertical',
                                      command=self.scroll)
        self.listbox = tk.Listbox(self, listvariable=self.items,
                                  width=kwargs.get('width', 45),
                                  height=kwargs.get('height', 7),
                                  yscrollcommand=self.scrollbar.set,
                                  exportselection=0)
        self.listbox.pack(side='left', fill='both', expand=True)
        self.scrollbar.pack(side='left', fill='y')
        #self.listbox.bind('<Button-1>', self.select)
        # use ButtonRelease so drag/drop functions don't mess with focus_set
        self.listbox.bind('<ButtonRelease-1>', self.select)
        self.listbox.bind('<Return>', self.next_item)
        # Callback methods for the various list-modification actions
        self.callbacks = {'insert': [], 'remove': [], 'select': [], 'swap': []}


    def set_variable(self, listvar):
        """Set the ScrollList to use the given ListVar for its items.
        """
        if not isinstance(listvar, ListVar):
            raise TypeError("ScrollList.set_variable requires a ListVar "
                            "(got %s instead)" % listvar.__class__.__name__)
        self.items = listvar
        self.listbox.config(listvariable=self.items)


    def scroll(self, *args):
        """Event handler when the list is scrolled.
        """
        self.listbox.yview(*args)
        # Scroll in the linked list also
        if self.linked:
            self.linked.listbox.yview(*args)


    def add(self, *values):
        """Add the given values to the end of the list.
        Calls ``insert`` callbacks for each added item.
        """
        for value in values:
            self.summon_callbacks('insert', self.items.count(), value)
            self.listbox.insert('end', value)
        self.select_index(-1)


    def insert(self, index, *values):
        """Insert values at a given index.
        Calls ``insert`` callbacks for each added item.
        """
        for offset, value in enumerate(values):
            self.listbox.insert(index+offset, value)
            self.summon_callbacks('insert', index+offset, value)
        self.select_index(index)


    def delete(self, first, last=None):
        """Delete values in a given index range (``first``, ``last``), not
        including ``last`` itself. If ``last`` is ``None``, delete only the
        item at index ``first``.  Calls ``remove`` callbacks for each removed item.
        """
        if first == last:
            return
        # Summon callbacks for each item/value about to be deleted
        for index in reversed(range(first, last or first+1)):
            value = self.items[index]
            self.summon_callbacks('remove', index, value)
        # Delete items from the listbox
        self.listbox.delete(first, last)
        self.select_index(first)


    def swap(self, index_a, index_b):
        """Swap the element at ``index_a`` with the one at ``index_b``.
        Calls ``swap`` callbacks for the swapped items.
        """
        # Use a temporary list, so self.items is only modified once
        temp_list = self.items.get()
        item_a = temp_list[index_a]
        item_b = temp_list[index_b]
        temp_list[index_a] = item_b
        temp_list[index_b] = item_a
        # Set the updated list
        self.items.set(temp_list)
        # Summon 'swap' callbacks
        self.summon_callbacks('swap', index_a, index_b)


    def select(self, event):
        """Event handler when an item in the list is selected.
        """
        index = self.listbox.nearest(event.y)
        self.select_index(index)


    def next_item(self, event):
        """Event handler when <Return> is pressed
        """
        self.curindex += 1
        if self.linked:
            self.linked.curindex = self.curindex
        self.select_index(self.curindex)


    def select_index(self, index, select_in_linked=True):
        """Select (highlight) the list item at the given ``index``.
        Calls ``select`` callbacks for the selected item.
        """
        item_count = self.items.count()
        if item_count == 0:
            return
        # If index is negative, or past the end, cycle or select last item
        if index < 0 or index >= item_count:
            # allow cycling (any unforeseen consequences ?)
            if self.listbox.get(index-1):
                index = 0
            else:
                index = item_count - 1
        # Clear selection, and select only the given index
        self.listbox.selection_clear(0, item_count)
        self.listbox.selection_set(index)
        # Set selected variable
        self.selected.set(self.listbox.get(index))
        self.curindex = index
        # Summon 'select' callbacks
        self.summon_callbacks('select', index, self.selected.get())
        # Select same index in linked ScrollList
        if self.linked and select_in_linked:
            self.linked.select_index(index, False)


    def get(self):
        """Return a list of all entries in the list.
        """
        return self.items.get()


    def set(self, values):
        """Set the list values to those given.
        """
        # Use delete / add so the relevant callbacks are summoned
        self.delete(0, self.items.count())
        self.add(*values)


    def link(self, scrolllist):
        """Link to another ScrollList, so they scroll and select in unison.
        """
        if not isinstance(scrolllist, ScrollList):
            raise TypeError("Can only link to a ScrollList.")
        self.linked = scrolllist
        scrolllist.linked = self


    def callback(self, action, function):
        """Add a callback function for the given action.

            action
                May be 'insert', 'remove', 'select', or 'swap'
            function
                Callback function to call for the given action

        All callback functions except 'swap' must take ``(index, value)``
        parameters; the 'swap' callback should accept ``(index_a, index_b)``.
        In all cases, ``index`` is 0-based, and -1 is passed for the last item.
        """
        if action not in ['insert', 'remove', 'select', 'swap']:
            raise ValueError("List callback action must be"
                             " 'insert', 'remove', 'select', or 'swap'.")
        if not hasattr(function, '__call__'):
            raise TypeError("List callback function must be callable.")
        self.callbacks[action].append(function)


    def summon_callbacks(self, action, index, item):
        """Summon callbacks for the given ``action``, passing ``index`` and
        ``item`` to each.
        """
        if action not in ['insert', 'remove', 'select', 'swap']:
            raise ValueError("Callback action must be"
                             " 'insert', 'remove', 'select', or 'swap'")
        for function in self.callbacks[action]:
            function(index, item)



class DragList (ScrollList):
    """A scrollable listbox with drag-and-drop support.
    """
    def __init__(self, master=None, items=None,
                 selected=None):
        """Create a DragList widget.

            master
                Tkinter widget that will contain the DragList
            items
                ListVar or Python list of items to show in listbox
            selected
                Tk StringVar to store currently selected list item
        """
        ScrollList.__init__(self, master, items, selected)
        # Add bindings for drag/drop
        self.listbox.bind('<Button-1>', self.select)
        self.listbox.bind('<B1-Motion>', self.drag)
        self.listbox.bind('<ButtonRelease-1>', self.drop)


    def select(self, event):
        """Event handler when an item in the list is selected.
        """
        # Set currently selected item and change the cursor to a double-arrow
        ScrollList.select(self, event)
        if self.linked:
            self.linked.curindex = self.curindex
        self.config(cursor="double_arrow")


    def drag(self, event):
        """Event handler when an item in the list is dragged.
        """
        # If item is dragged to a new location, swap
        loc = self.listbox.nearest(event.y)
        if loc != self.curindex and self.items.count() > 0:
            self.swap(self.curindex, loc)
            self.curindex = loc


    def drop(self, event):
        """Event handler when an item in the list is "dropped".
        """
        # Change the mouse cursor back to the default arrow.
        self.config(cursor="")
        # make sure drop doesn't mess with focus_set
        self.select_index(self.curindex)


class ComboBox (tk.Frame):
    def __init__(self, master, choices=None,
                 variable=None, command=None,
                 **kwargs):
        """Create a ComboBox.

            master
                Tk Widget that will contain the ComboBox
            choices
                ListVar or Python list of available choices
            variable
                Tk StringVar to store currently selected choice in
            command
                Function to call when an item in the list is selected
        """
        tk.Frame.__init__(self, master)
        self.variable = variable or tk.StringVar()
        # Create and pack the Tix ComboBox
        self.combo = Tix.ComboBox(self, editable=True, **kwargs)
        self.combo.config(variable=self.variable, command=command)
        self.combo.pack(side='left', expand=True, fill='both')
        # Make the entry box background white
        # FIXME: This doesn't do anything when editable=False
        self.combo.subwidget('entry').config(bg='#ffffff')
        # Add choices, if specified
        if choices:
            for index, choice in enumerate(choices):
                self.combo.insert(index, choice)


class FontChooser (Dialog):
    """A widget for choosing an ImageMagick font name.
    """
    # Cache of PhotoImage previews, indexed by font name
    _cache = {}

    def __init__(self, parent=None):
        Dialog.__init__(self, parent, "ImageMagick font chooser")
        # Defined in body()
        self.fontlist = None
        self.preview = None


    def body(self, master):
        """Draw widgets inside the Dialog, and return the widget that should
        have the initial focus. Called by the Dialog base class constructor.
        """
        tk.Label(master, text="Available fonts").pack(side='top')

        # List of fonts available to ImageMagick
        available_fonts = imagemagick_fonts()
        self.fontlist = ScrollList(master, available_fonts)
        self.fontlist.callback('select', self.refresh)
        self.fontlist.pack(side='top', fill='both', expand=True)

        # Font preview area
        self.preview = tk.Label(master, image=None, width=500)
        self.preview.pack(fill='both', expand=True)

        # Draw the initial preview
        self.refresh()

        # Return widget with initial focus
        return self.fontlist


    def apply(self):
        """Set the selected font. Called by base Dialog when "OK" is pressed.
        """
        self.result = self.fontlist.selected.get()


    def refresh(self, index=0, fontname='Helvetica'):
        """Redraw the preview using the current font and size.
        """
        # Cache a PhotoImage for this font name
        if fontname not in self._cache:
            self._cache[fontname] = self.render_font(fontname)
        # Get the PhotoImage from the cache
        photo_image = self._cache[fontname]
        self.preview.configure(image=photo_image, height=photo_image.height())


    def render_font(self, fontname):
        """Return a `tk.PhotoImage` preview of the given font.
        """
        cmd = cli.Command('convert')
        cmd.add('-size',  '500x60')
        cmd.add("xc:#EFEFEF")
        cmd.add('-font', fontname, '-pointsize', 24)
        cmd.add('-gravity', 'center', '-annotate', '+0+0', fontname)
        cmd.add('gif:-')
        cmd.run(capture=True)
        image_data = cmd.get_output()
        return tk.PhotoImage(data=base64.b64encode(image_data))


class PopupScale (Dialog):
    """A popup widget with a scale for choosing a number.
    """
    def __init__(self, parent=None):
        """Create the scale dialog with the given Number control as a parent.
        """
        Dialog.__init__(self, parent, parent.label or 'Scale')
        # Set by body()
        self.scale = None


    def body(self, master):
        """Draw the popup dialog and return the widget that should have the
        initial focus.
        """
        num = self.parent
        frame = tk.Frame(self, relief=tk.GROOVE, borderwidth=3)
        frame.pack(side='top', fill='both', expand=True, pady=6)
        self.scale = tk.Scale(frame, from_=num.min, to=num.max,
                               resolution=num.step,
                               tickinterval=(num.max - num.min),
                               orient='horizontal')
        self.scale.pack(side='left', fill='both', expand=True)
        tk.Label(frame, name='units', text=num.units).pack(side='left')
        # Set scale to parent Number's value initially
        self.scale.set(num.variable.get())


    def apply(self):
        """Set the result to the current scale value. Called when "OK" is
        pressed.
        """
        self.result = self.scale.get()


class Style:
    """Generic widget style definitions.
    """
    def __init__(self,
                 bgcolor='#ffffff',
                 fgcolor='#d9d9d9',
                 textcolor='#000000',
                 font=('Helvetica', 10, 'normal'),
                 relief='groove'):
        self.bgcolor = bgcolor
        self.fgcolor = fgcolor
        self.textcolor = textcolor
        self.font = font
        self.relief = relief


    def apply(self, root):
        """Apply the current style to the given Tkinter root window.
        """
        assert isinstance(root, tk.Tk)
        root.option_clear()
        # Font
        root.option_add("*font", self.font)
        #root.option_add("*Text.font", ('Courier', 10, 'normal'))
        root.option_add("*Text.font", self.font)
        # Background color
        root.option_add("*Scale.troughColor", self.bgcolor)
        root.option_add("*Spinbox.background", self.bgcolor)
        root.option_add("*Entry.background", self.bgcolor)
        root.option_add("*Listbox.background", self.bgcolor)
        root.option_add("*Text.background", self.bgcolor)
        # Button colors
        root.option_add("*Radiobutton.selectColor", "#ffffff")
        root.option_add("*Checkbutton.selectColor", "#ffffff")
        # Relief
        root.option_add("*Entry.relief", self.relief)
        root.option_add("*Spinbox.relief", self.relief)
        root.option_add("*Listbox.relief", self.relief)
        root.option_add("*Button.relief", self.relief)
        root.option_add("*Menu.relief", self.relief)
        # Mouse-over effects
        root.option_add("*Button.overRelief", 'raised')
        root.option_add("*Checkbutton.overRelief", 'raised')
        root.option_add("*Radiobutton.overRelief", 'raised')


    def save(self, filename):
        """Save the current style settings to an .ini-formatted config file.
        """
        # Load existing config file
        config = ConfigParser()
        config.read(filename)

        # Save font settings
        if 'font' not in config.sections():
            config.add_section('font')
        family, size, style = self.font
        config.set('font', 'family', family)
        config.set('font', 'size', size)
        config.set('font', 'style', style)

        # TODO: Save other style settings

        # Yuck...
        dirname, fname = os.path.split(filename)
        if not os.path.exists(dirname):
            os.mkdir(dirname)
        outfile = open(filename, 'w')
        config.write(outfile)
        outfile.close()


    def load(self, filename):
        """Load style settings from an .ini-formatted config file.
        """
        config = ConfigParser()
        config.read(filename)
        font = dict(config.items('font'))
        self.font = (font['family'], int(font['size']), font['style'])


class ConfigWindow (Dialog):
    """Configuration settings dialog box.
    """
    def __init__(self, master=None, style=None):
        """Create and display a configuration window.

            inifile
                An ``.ini``-style file to load/save settings from
        """
        self.style = style or Style()
        Dialog.__init__(self, master, "Configuration")
        # Defined in body()
        self.fontfamily = None
        self.fontsize = None
        self.fontstyle = None


    def body(self, master):
        """Draw widgets inside the Dialog, and return the widget that should
        have the initial focus. Called by the Dialog base class constructor.
        """
        # Font family
        families = sorted(tkFont.families())
        tk.Label(master, text="Font family").pack(side='top')
        self.fontfamily = ScrollList(master, families)
        self.fontfamily.callback('select', self.refresh)
        self.fontfamily.pack(side='top', fill='both', expand=True)
        # Font style
        tk.Label(master, text="Font style").pack(side='top')
        self.fontstyle = ScrollList(master, ['normal', 'bold'], height=2)
        self.fontstyle.callback('select', self.refresh)
        self.fontstyle.pack(side='top')
        # Font size
        tk.Label(master, text="Font size").pack(side='top')
        size_var = tk.StringVar()
        self.fontsize = tk.Spinbox(master, from_=8, to=24,
           textvariable=size_var, command=self.refresh)
        self.fontsize.variable = size_var
        self.fontsize.pack(side='top')
        # Preview area
        preview_frame = tk.LabelFrame(master, text="Preview")
        preview_frame.pack(side='top', fill='both')
        self.preview = tk.Label(preview_frame,
            text="The quick brown fox")
        self.preview.pack(side='top')
        # Use initial values loaded from .ini file
        family, size, style = self.style.font
        self.fontfamily.selected.set(family)
        self.fontsize.variable.set(size)
        self.fontstyle.selected.set(style)
        # Refresh the preview
        self.refresh()
        # Return widget with initial focus
        return self.fontfamily


    def get_font(self):
        """Return a ``(family, size, style)`` tuple for the current font.
        """
        return (self.fontfamily.selected.get(),
                self.fontsize.variable.get(),
                self.fontstyle.selected.get())


    def refresh(self, *args):
        """Refresh the preview to show the current font.
        """
        self.preview.config(font=self.get_font())


    def apply(self):
        """Apply the selected configuration settings.
        """
        self.style.font = self.get_font()
        self.result = self.style


class ScrolledWindow (tk.Tk):
    """A top-level window with scrollbars.

    To use as a container for other widgets, do::

        window = ScrolledWindow()
        button = tk.Button(window.frame, text="Click me", ...)
        entry = tk.Entry(window.frame, ...)
        window.draw()

    That is, use window.frame as the master of child widgets, instead of
    window itself. (TODO: Eliminate this requirement.)
    """
    def __init__(self, width, height):
        tk.Tk.__init__(self)
        self.width = width
        self.height = height
        # Frame inside canvas, fills all available space
        self.canvas = tk.Canvas(self, width=self.width, height=self.height)
        self.frame = tk.Frame(self.canvas)
        # Grid fills all available window space
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)


    def draw(self):
        """Draw the scrollbars and container frame.
        """
        # Put the container frame in the Canvas
        self.canvas.create_window(0, 0, window=self.frame, anchor='nw')
        # Canvas scrollable area
        self.canvas.configure(scrollregion=(0, 0, self.width, self.height))
        self.canvas.grid(row=0, column=0, sticky='nsew')
        # Attach scrollbars to the Canvas
        h_scroll = tk.Scrollbar(self, orient='horizontal',
                                command=self.canvas.xview)
        v_scroll = tk.Scrollbar(self, orient='vertical',
                                command=self.canvas.yview)
        h_scroll.grid(row=1, column=0, sticky='we')
        v_scroll.grid(row=0, column=1, sticky='ns')
        self.canvas.configure(xscrollcommand=h_scroll.set,
                              yscrollcommand=v_scroll.set)


class PrettyLabel (tk.Text):
    """Like a Label, with automatic word-wrapping of paragraphs, and inference
    of headings and preformatted text from basic markup cues.

    """
    def __init__(self, master, text, font=None, width=65, height=24):
        """Create a pretty label for displaying the given text.
        """
        # Text widget with the same background color as the master
        tk.Text.__init__(self, master, wrap='word', borderwidth=0,
           padx=10, width=width, height=height, exportselection=0)
        if font:
            self.config(font=font)
        # Add the text with formatting applied
        self.set_text(text)
        # background and unfocused border color same as master's background
        bg = master.cget('background')
        self.config(background=bg, highlightbackground=bg)
        # Make the text widget read-only
        self.config(state='disabled')


    def rewrap(self, text):
        """Rewrap the given text, by joining regular paragraphs into
        single lines, and leaving preformatted text and headings alone.
        """
        # Start by splitting into chunks at double-line-breaks
        lines = []
        chunks = text.strip().split('\n\n')
        for chunk in chunks:
            # If this chunk contains newlines, they could be in paragraphs
            # (which should wrap) or preformatted text (which should not be
            # wrapped)
            if '\n' in chunk:
                parts = chunk.split('\n')
                # Preformatted - append each line individually
                if all(line.startswith(' ') for line in parts):
                    lines.extend(parts)
                # Paragraph - join lines with spaces
                else:
                    lines.append(' '.join(parts))
            # No newline in chunk -- It's either a heading or a
            # one-line paragraph, so append it as-is
            else:
                lines.append(chunk)
            # Include an empty line where the \n\n was before
            lines.append('')
        return lines


    def set_text(self, text):
        """Set the widget to contain the given text, with formatting applied.
        """
        heading_re = re.compile('^[ A-Z()]+$')
        pre_re = re.compile('^ +(.+)$')

        # Parse each logical (wrapped) line in the text
        lineno = 1
        for line in self.rewrap(text):
            L = str(lineno)
            # Heading line
            if heading_re.match(line):
                # Capitalize each word in the heading
                heading = ' '.join(word.capitalize() for word in line.split(' '))
                self.insert('end', heading)
                # Unique tag name for this heading
                tag = 'heading' + L
                self.tag_add(tag, L+'.0', L+'.end')
                self.tag_config(tag, font=self.get_heading_font())
            # Preformatted line
            elif pre_re.match(line):
                self.insert('end', line)
                # Unique tag name for this line
                tag = 'pre' + L
                self.tag_add(tag, L+'.0', L+'.end')
                self.tag_config(tag, font=self.get_monospace_font())
            # Paragraph
            else:
                self.insert('end', line)
            # Newline to separate each logical line
            self.insert('end', '\n')
            lineno += 1


    def get_heading_font(self):
        """Return a bold font, in the current font family.
        """
        # The config('font') method returns a 5-tuple, where the last
        # item is the font descriptor tuple, or 'TkFixedFont' as a default
        font = self.config('font')[-1]
        # If font is (family, height, style), increase size and make it bold
        if isinstance(font, tuple) and len(font) == 3:
            family, height, style = font
            return (family, int(height)+6, 'bold')
        # Default font
        else:
            return ('Helvetica', 16, 'bold')


    def get_monospace_font(self):
        """Return a monospace font, with the current font size and style.
        """
        font = self.config('font')[-1]
        # If font is (family, height, style), change family to Courier
        if isinstance(font, tuple) and len(font) == 3:
            family, height, style = font
            return ('Courier', int(height), style)
        # Default font
        else:
            return ('Courier', 12, 'normal')




# -*- coding: utf8 -*-

# Copyright 2007 Joe Friedrichsen <pengi.films@gmail.com>
#
# This file is part of tovid.

"""Relate a video's bitrate, size, and play time

A video file (a/v stream) has three related characteristics:

    - Play length
    - Final output size
    - Encoded (average) bitrate

These are related by their units: ``bitrate = size / length``. This module
provides ways to calculate these values.

You can predict/calculate any one of these characteristics given the
other two. By default, a new AVstream object assumes you want to find
the average bitrate from a known play length and final size::

    >>> avs = playtime.AVstream()
    >>> avs.play_length
    120.0
    >>> avs.final_size
    4400.0
    >>> avs.bitrate.kbps
    5126.3715555555555
    >>> avs.bitrate.MiBpm
    36.666666666666664
    >>> avs.bitrate.GiBph
    2.1484375

Usually when putting video on a disc, the final output size is well
defined and non-changing. By default, AVstream fixes this characteristic so
that you can see how the bitrate changes for different amounts of time on
that disc::

    >>> avs.set_play_length(180)
    >>> avs.bitrate.kbps
    3417.5810370370368

However, if you know the video is a certain length, you can fix that instead
and find how bitrates change according to different output sizes::

    >>> avs.set_fixed_param('LENGTH')
    >>> avs.set_final_size(2000)
    >>> avs.bitrate.kbps
    1537.9114666666667

Finally, you're not limited to finding bitrates. You can fix the bitrate and
see how output size and play length are related::

    >>> avs.set_bitrate(1152, 'kbps')
    >>> avs.set_fixed_param('RATE')
    >>> avs.set_final_size(700)
    >>> avs.play_length
    84.954074074074072
    >>> avs.set_play_length(120)
    >>> avs.final_size
    988.76953125

"""

class AVstream:
    """Video file bitrate/size/length calculator object

    An AVstream object lets you calculate a video's bitrate, final size, or
    play length given the other two. Usually, you want to find the bitrate,
    and this is what AVstream does by default.

    Once instantiated, the bitrate can be accessed in three different units:

        - AVstream.bitrate.kbps -- kilobits per second (conventional)
        - AVstream.bitrate.MiBpm -- Mibibytes per minute
        - AVstream.bitrate.GiBph -- Gibibytes per hour

    There are four attributes:

        play_length
            the length of the video in minutes
        final_size
            the size of the video in MiB
        bitrate
            the bitrate of the video
        fixed_param
            the fixed characteristic

    Each of these attributes have a 'set_NAME' method that should be used
    to make changes so that the other attributes are updated automatically.

    """
    def __init__(self, play_length=120.0, final_size=4400.0):
        """Create a new AVstream object

            play_length
                the length in minutes (default = 120.0)
            final_size
                the final size in MiB (default = 4400.0)

        """
        self.play_length = play_length
        self.final_size = final_size
        self.bitrate = Bitrate( (final_size/play_length), 'MiBpm')
        self.fixed_param = "SIZE"

    def set_bitrate(self, bitrate, units):
        """Set the bitrate for the stream and recalculate the variables
        according to the fixed parameter.

            bitrate
                number (integer or float ok)
            units
                the units that the bitrate is in. Valid unit arguments are
                'kbps', 'MiBpm', or 'GiBph'

        """
        self.bitrate.set(bitrate, units)
        if self.fixed_param == "RATE":
            pass
        elif self.fixed_param == "LENGTH":
            self._calculate_final_size()
        elif self.fixed_param == "SIZE":
            self._calculate_play_length()

    def set_play_length(self, play_length):
        """Set the play length in minutes and recalculate variables according
        to the fixed parameter.

            play_length
                how long the video is (minutes)

        """
        self.play_length = play_length
        if self.fixed_param == "RATE":
            self._calculate_final_size()
        elif self.fixed_param == "LENGTH":
            pass
        elif self.fixed_param == "SIZE":
            self._calculate_bitrate()

    def set_final_size(self, final_size):
        """Set the final size in MiB (Mebibytes) and recalculate variables
        according to the fixed parameter.

            final_size
                how large the final size can/should be (MiB)

        """
        self.final_size = final_size
        if self.fixed_param == "RATE":
            self._calculate_play_length()
        elif self.fixed_param == "LENGTH":
            self._calculate_bitrate()
        elif self.fixed_param == "SIZE":
            pass

    def set_fixed_param(self, param):
        """Set the fixed parameter of the AVstream object.

            param
                The parameter to fix. Valid arguments are:
                - RATE (the bitrate of the AVstream)
                - LENGTH (the play length of the AVstream)
                - SIZE (the final size of the AVstream)

        """
        valid_params = ["RATE", "SIZE", "LENGTH"]
        if param in valid_params:
            self.fixed_param = param
        else:
            print("%s: bad new fixed_param -- %s" % ('playtime', param))
            print("%s: keeping old fixed_param -- %s" % \
                ('playtime', self.fixed_param))

    def _calculate_bitrate(self):
        """Find the bitrate given the length and size"""
        self.bitrate.set( (self.final_size/self.play_length), 'MiBpm')

    def _calculate_final_size(self):
        """Find the final size give the bitrate and length"""
        self.final_size = self.bitrate.MiBpm * self.play_length

    def _calculate_play_length(self):
        """Find the length given the bitrate and size"""
        self.play_length =  self.final_size / self.bitrate.MiBpm


class Bitrate:
    """Convert between different bitrate units

    A Bitrate object stores a bitrate in three different units:

        kbps
            kilobits per second (the conventional unit)
        MiBpm
            Mibibytes per minute (uses for S/VCD sizes)
        GiBph
            Gibibytes per hour (useful for DVD sizes)

    Access these by name::

        >>> br = playtime.Bitrate(3300)
        >>> br.kbps
        3300
        >>> br.MiBpm
        23.603439331054688
        >>> br.GiBph
        1.3830140233039856

    Once instantiated or set with a given bitrate and unit, the other
    remaining bitrates are automatically calculated and updated::

        >>> br.set(1, 'GiBph')
        >>> br.kbps
        2386.0929422222221

    """
    def __init__(self, bitrate, unit='kbps'):
        """Create a new Bitrate object

            bitrate
                the bitrate (int or float ok)
            units
                'kbps', 'MiBpm', or 'GiBph'
        """
        self.set(bitrate, unit)

    def set(self, bitrate, unit):
        """Set the bitrate in 'unit' units.

            bitrate
                new bitrate (integer or float ok)
            units
                'kbps', 'MiBpm', or 'GiBph'

        Once set with a given bitrate and unit, the other remaining
        bitrates are automatically calculated.

        """
        valid_units = ["kbps", "MiBpm", "GiBph"]
        if unit in valid_units:
            self.unit = unit
            if self.unit == 'kbps':
                self.kbps = bitrate
                self._to_MiBpm()
                self._to_GiBph()
            elif self.unit == 'MiBpm':
                self.MiBpm = bitrate
                self._to_kbps()
                self._to_GiBph()
            elif self.unit == 'GiBph':
                self.GiBph = bitrate
                self._to_kbps()
                self._to_MiBpm()
        else:
            print("%s: bad new units -- %s" % ('playtime', unit))
            print("%s: keeping old units -- %s" % ('playtime', self.unit))
            print("%s: keeping old bitrate -- %s" % \
                ('playtime', getattr(self, self.unit)))

    def _to_kbps(self):
        """Convert the bitrate to kbps"""
        if self.unit == 'kbps':
            bitrate = self.kbps
        elif self.unit == 'MiBpm':
            bitrate = self.MiBpm * (8.0*1024.0*1024.0) / (60.0*1000.0)
        elif self.unit == 'GiBph':
            bitrate = self.GiBph * (8.0*1024.0*1024.0*1024.0) / (60.0*60.0*1000.0)
        self.unit = 'kbps'
        self.kbps = bitrate

    def _to_MiBpm(self):
        """Convert the bitrate to MiBpm"""
        if self.unit == 'kbps':
            bitrate = self.kbps * (60.0*1000.0) / (8.0*1024.0*1024.0)
        elif self.unit == 'MiBpm':
            bitrate = self.MiBpm
        elif self.unit == 'GiBph':
            bitrate = self.GiBph * 1024.0 / 60.0
        self.unit = 'MiBpm'
        self.MiBpm = bitrate

    def _to_GiBph(self):
        """Convert the bitrate to GiBph"""
        if self.unit == 'kbps':
            bitrate = self.kbps * (60.0*60.0*1000.0) / (8.0*1024.0*1024.0*1024.0)
        elif self.unit == 'MiBpm':
            bitrate = self.MiBpm * 60.0 / 1024.0
        elif self.unit == 'GiBph':
            bitrate = self.GiBph
        self.unit = 'GiBph'
        self.GiBph = bitrate




# tovid.py

"""A GUI for the ``tovid`` script.
"""

# Get supporting classes from libtovid.metagui
from libtovid.metagui import *

### --------------------------------------------------------------------
### tovid
### --------------------------------------------------------------------

_in = Filename('Input filename', '-in', '',
    'Video file to encode',
    'load', 'Select a video file')
_out = Filename('Output prefix', '-out', '',
    'Name to use for encoded output file (.mpg added automatically)',
    'save', 'Choose an output prefix')

# Standard formats
_dvd = Flag("DVD", '-dvd', False,
    "(720x480 NTSC, 720x576 PAL) DVD-compatible output. May be burned "
    "to a DVD[+/-]R[W] disc. Also known as Digital [Versatile|Video] Disc, "
    "or just DVD depending on who you talk to.")
_svcd = Flag("Super Video CD (SVCD)", '-svcd', False,
    "(480x480 NTSC, 480x576 PAL) Super Video CD format, may be burned "
    "to a CD-R. Like VCD but with better resolution and variable bitrate. "
    " About 1 hour of playing time per disc.")
_vcd = Flag("Video CD (VCD)", '-vcd', False,
    "(320x240 NTSC, 320x288 PAL) Video CD format, may be burned to a CD-R. "
    "About 1 hour of playing time per disc.")

# Extra formats
_dvd_vcd = Flag('VCD-on-DVD', '-dvd-vcd', False,
    '(352x240 NTSC, 352x288 PAL) VCD-on-DVD output')
_half_dvd = Flag('Half-DVD', '-half-dvd', False,
    '(352x480 NTSC, 352x576 PAL) Half-D1-compatible output')
_kvcd = Flag('KVCD', '-kvcd', False,
    '(352x240 NTSC, 352x288 PAL) KVCD-enhanced long-playing video CD')
_kvcdx3 = Flag('KVCDx3', '-kvcdx3', False,
    '(528x480 NTSC, 520x576 PAL) KVCDx3 specification')
_kvcdx3a = Flag('KVCDx3A', '-kvcdx3a', False,
    '(544x480 NTSC, 544x576 PAL) KVCDx3a specification')
_kdvd = Flag('KDVD', '-kdvd', False,
    '(720x480 NTSC, 720x576 PAL) KVCD-enhanced long-playing DVD')
_bdvd = Flag('BDVD', '-bdvd', False,
    '(720x480 NTSC, 720x576 PAL) BVCD-enhanced long-playing DVD')

# Aspect ratio
_full = Flag('4:3 (full-frame)', '-full', False)
_wide = Flag('16:9 (widescreen)', '-wide', False)
_panavision = Flag('2.35:1 (panavision)', '-panavision', False)
_aspect = Text('Aspect ratio', '-aspect', '4:3',
    'Explicit integer aspect ratio (ex. 6:4)')

# TV Systems
_ntsc = Flag("NTSC", '-ntsc', False, "NTSC, US standard, 29.97 fps")
_ntscfilm = Flag("NTSC Film", '-ntscfilm', False, "NTSC Film, 23.976 fps")
_pal = Flag("PAL", '-pal', False, "PAL, European standard, 25.00 fps")

# Video quality/size controls
_quality = Number('Quality', '-quality', 6,
    'Quality of encoding, on a scale of 1-10, with 10 giving the best '
    'quality, at the expense of a larger output file.',
    1, 10)
_vbitrate = Number('Bitrate', '-vbitrate', 0,
    'Target video bit rate, in kilobits per second. Higher bit rates '
    'give better quality, but a larger output file.',
    0, 9000, units='kbits/sec')
_fit = Number('Fit to', '-fit', 0,
    'Fit the output file into the given size in MiB (2^20 bytes). '
    'Video bitrate and quantization are chosen to ensure the target size '
    'is not exceeded. Ignored for VCD, which has constant bitrate.',
    0, 4400, units='MiB')
_discsize = Number('Disc size', '-discsize', 0,
    'Size of target media, in MiB (2^20 bytes). Output will be split into '
    'chunks of this size.',
    0, 4400, units='MiB')

# Encoder options
_mplayeropts = Text('mplayer options', '-mplayeropts', '', 'TODO: Tooltip')
_filters = Choice('mplayer filters', '-filters', 'none', 'TODO: Tooltip',
    'none|denoise|deblock|contrast|all', side='top')
_ffmpeg = Flag('Encode using ffmpeg', '-ffmpeg', False)
_parallel = Flag('Parallel video/audio encoding', '-parallel', False,
    "(mpeg2enc only) Rip, encode, and multiplex in parallel.")

# Interlacing
_interlaced = Flag('Interlaced encoding', '-interlaced', False,
    "Do interlaced encoding. Use this if your source material is interlaced.")
_interlaced_bf = Flag('Interlaced (bottom first)', '-interlaced_bf', False,
    "Do interlaced encoding, bottom-field first. Use this if "
    "the -interlaced option produces bad output.")
_deinterlace = Flag('Deinterlace', '-deinterlace', False,
    "Deinterlace the input video, and create a progressive-scan output. "
    "Degrades video quality considerably.")

# Picture manipulation
_safe = Number('Safe area', '-safe', 90,
    'Safe area as a percentage of screen size',
    50, 100, 'scale')
_crop = Text('Crop', '-crop', '', 'TODO: Tooltip')
_slice = Number('Slice', '-slice', 0, 'TODO: Tooltip', 0, 220000)
_fps = Text('FPS', '-fps', '', 'TODO: Tooltip')
_type = Choice('Video type', '-type', 'live', 'TODO: Tooltip',
    'live|animation|bw', side='top')

# Subtitles
_autosubs = Flag('Auto-subtitles', '-autosubs', False,
    'Automatically include subtitle files with the same '
    'name as the input video.')
_mkvsub = Text('mkvsub', '-mkvsub', '',
    'EXPERIMENTAL. Attempt to encode an integrated subtitle stream '
    '(such as may be found in Matroska .mkv files) in the given '
    'language code (eng, jpn, etc.) May work for other formats.')
_subtitles = Filename('subtitles', '-subtitles', '',
    'Get subtitles from FILE and encode them into the video. '
    'WARNING: This hard-codes the subtitles into the video, and you '
    'cannot turn them off while viewing the video. By default, no subtitles '
    'are loaded. If your video is already compliant with the chosen output '
    'format, it will be re-encoded to include the subtitles.',
    'load', 'Select a subtitle file')

# Audio options
_normalize = Flag('Normalize', '-normalize', False,
    'Analyze the audio stream and then normalize the volume of the audio. '
    'This is useful if the audio is too quiet or too loud, or you want to '
    'make volume consistent for a bunch of videos. Similar to running '
    'normalize without any parameters. The default is -12dB average level '
    'with 0dB gain.')
_downmix = Flag('Downmix', '-downmix', False,
    'Encode all audio tracks as stereo. This can save space on your DVD if '
    'your player only does stereo. The default behavior of tovid is to use the '
    'original number of channels in each track. For aac audio, downmixing is '
    'not possible: tovid runs a quick 1 frame test to try to downmix the input '
    'track with the largest number of channels, and if it fails then it will '
    'revert to the default behavior of using the original channels.')
_audiotrack = SpacedText('Audio track', '-audiotrack', '',
    'Encode  the  given  audio  track, if the input video has multiple '
    'audio tracks.  NUM is 1 for the first track, 2 for the second, etc. '
    'You may also provide a list of tracks, separated by spaces or commas, '
    'for example -audiotrack 3,1,2. Use idvid on your source video to '
    'determine which audio tracks it contains. ')
_abitrate = Number('Audio bitrate', '-abitrate', 224,
   'Encode audio at NUM kilobits per second. Reasonable values include 128, '
    '224,  and  384.  The default is 224 kbits/sec, good enough for most '
    'encodings. The value must be within the allowable range for the chosen '
    'disc format; Ignored for VCD, which must be 224.',
    128, 384, units='kbits/sec')
_amplitude = Text('Amplitude', '-amplitude', '',
    'In addition to analyzing and normalizing, apply the gain to the audio '
    'such that the "average" (RMS) sound level is NUM. Valid values range '
    '0.0 - 1.0, with  0.0  being  silent and 1.0 being full scale. Use NUMdB '
    'for a decibel gain below full scale '
    '(the default without -amplitude is -12dB).')
_async = Number('Audio sync', '-async', 0, 'Adjust audio synchronization',
    -600, 600, units='secs')

# File-related
_keepfiles = Flag('Keep temporary files', '-keepfiles')
_overwrite = Flag('Overwrite files', '-overwrite')
_nofifo = Flag('No FIFO', '-nofifo', False,
    "Don't use a FIFO (pipe) when encoding. Instead, write "
    "intermediate conversion files to disk. May take "
    "up lots of space!")

# Runtime behavior
_config = Filename('Configuration file', '-config', '', 'TODO: Tooltip',
    'load', 'Select tovid configuration file')
_priority = Choice('Process priority', '-priority', 'high', 'TODO: Tooltip',
    'low|medium|high')
_update = Number('Update interval', '-update', 5, 'TODO: Tooltip', 1, 20)
_quiet = Flag('Quiet', '-quiet')
_fake = Flag('Fake encoding', '-fake')
_force = Flag('Force encoding', '-force')

# Prompting
_noask = Flag('No prompts',  '-noask', False, 'TODO: Tooltip')
_from_gui = Flag('From GUI', '-from-gui', True)

### ---------------------------------------------------------------------------
### Higher-level groups and panels
### ---------------------------------------------------------------------------

IN_OUT = VPanel('Filenames', _in, _out)
FORMAT = FlagGroup('Disc format', 'exclusive',
    _dvd, _half_dvd, _kvcd,
    _kvcdx3a, _kdvd, _bdvd,
    side='top',
    columns=3)

BASIC_OPTS = VPanel('Basic options',
    FORMAT,
    HPanel('Output file size', _quality, _vbitrate, _fit, _discsize),
    HPanel('',
        FlagGroup('TV System', 'exclusive', _ntsc, _ntscfilm, _pal),
        FlagGroup('Aspect ratio', 'exclusive', _full, _wide, _panavision),
        FlagGroup('Interlacing', 'exclusive',
                  _interlaced, _interlaced_bf, _deinterlace)
    ),
)

MAIN = VPanel('Main',
    IN_OUT,
    BASIC_OPTS,
)

VIDEO = HPanel('Video',
    VPanel('Encoder options',
        FlagGroup('', 'exclusive', _ffmpeg, _parallel),
        _filters,
        _mplayeropts,
        _force,
    ),
    VPanel('Picture manipulation',
        _safe, _crop, _slice, _fps, _type),
)

AUDIO = VPanel('Audio & Subtitles',
    _normalize,
    _downmix,
    _audiotrack,
    _abitrate,
    _amplitude,
    _async,
    _autosubs,
    _mkvsub,
    _subtitles,
)

BEHAVIOR = VPanel('Behavior',
    _config,
    _priority,
    VPanel('File output', _fake, _overwrite, _keepfiles, _nofifo),
    VPanel('Prompts and output options', _quiet, _noask, _from_gui, _update)
)

### --------------------------------------------------------------------
### main GUI
### --------------------------------------------------------------------

def run():
    app = Application('tovid', MAIN, VIDEO, AUDIO, BEHAVIOR)
    gui = GUI("tovid metagui", 640, 720, app)
    gui.run()

if __name__ == '__main__':
    run()




import time
import shlex
import re
import os
import fnmatch
from libtovid.metagui import *
from libtovid.metagui.control import _SubList
from libtovid.util import filetypes
from subprocess import Popen, PIPE
from tempfile import mkdtemp, mkstemp
from sys import stdout

try:
    from commands import getstatusoutput, getoutput
    import Tkinter as tk
except ImportError:
    # python 3
    from subprocess import getstatusoutput, getoutput
    import tkinter as tk

__all__ = [ 'VideoGui', 'SetChapters', 'Chapters', 'strip_all', 'to_title',
'find_masks', 'nodupes', 'video_filetypes', 'image_filetypes',
'visual_filetypes', 'dvd_video_files', 'av_filetypes', 'sys_dir',
'thumb_masks', 'home_dir', 'tovid_prefix', 'tovid_icon', 'os_path',
'heading_text', '_files_and_titles', '_out', 'CopyableInfo' ]


class VideoGui(tk.Frame):
    """A basic GUI to play video files.  It runs mplayer in slave mode
    so that commands can be sent to it via fifo.

    Without subclassing it only contains a 'play/pause button
    and an 'exit' button.
    """
    def __init__(self, master, args='', title='', callback=None):
        """Initialize GUI

           master
               widget that will conain this GUI
           args
               additional arguments to mplayer command, inserted
               at the end of the command just before the 'FILE' argument.
           title
               the wm title given to the master widget.
           callback
               a function run at program exit, before cleanup of temp files
        """
        tk.Frame.__init__(self, master)

        self.args = args
        self.show_osd = False
        if '-osdlevel 3' in args:
            self.show_osd = True
        self.master = master
        if title:
            try:
                self.master.title(title)
            except AttributeError:
                print("Error: " + \
                  "VideoGui master must be a root window for 'title' option")
        self.callback = callback
        self.v_width = 540
        self.v_height = 405
        self.is_running = tk.BooleanVar()
        self.is_running.set(False)
        self.pauseplay = tk.StringVar()
        self.pauseplay.set('Play')
        # temporary directory for fifo and other mplayer files
        self.make_tmps()
        self.draw()

    def make_tmps(self):
        """Make temporary directory containing fifo for mplayer commmands,
        editlist, and log
        """
        self.tempdir = mkdtemp(prefix='tovid-')
        self.cmd_pipe = os.path.join(self.tempdir, 'slave.fifo')
        self.editlist = os.path.join(self.tempdir, 'editlist')
        os.mkfifo(self.cmd_pipe)
        self.log = os.path.join(self.tempdir, 'mplayer.log')
        
    def draw(self):
        """Draw the GUI in self.master and get X11 identifier for container"""
        self.root_frame = tk.Frame(self)
        self.root_frame.pack(side='top', fill='both', expand=1, pady=20)
        self.container = tk.Frame(self.root_frame, bg='black', container=1, colormap='new')
        self.container.configure(width=self.v_width, height=self.v_height, bg='black')
        self.container.pack()
        # X11 identifier for the container frame
        self.xid = self.tk.call('winfo', 'id', self.container)
        self.add_widgets()

    def add_widgets(self):
        """Add buttons to the VideoGui.  Override this to customize buttons.
        root_frame has 'grab_set()' applied to it so make sure widgets go
        into this frame or they will not be functional!
        This function is called in draw()
        """
        button_frame = tk.Frame(self.root_frame)
        button_frame.pack(side='bottom', fill='x', expand=1)
        self.control_frame = tk.Frame(button_frame, borderwidth=1, relief='groove')
        self.control_frame.pack()
        self.load_button = tk.Button(self.control_frame,
            command=self.load, text='load video')
        self.load_button.pack(side='left')
        exit_button = tk.Button(self.control_frame, command=self.exit_mplayer, text='exit')
        self.pause_button = tk.Button(self.control_frame, command=self.pause,
                          width=12, textvariable=self.pauseplay)
        self.pause_button.pack(side='left')
        exit_button.pack(side='left', padx=5)
        self.mp_ctrls = [self.pause_button]
        self.toggle_controls('disabled', self.mp_ctrls)


    def identify(self, video):
        """Get information about video from mplayer -identify.
        Called by set_container()
        """
        
        cmd = 'mplayer -vo null -ao null -frames 5 -channels 6 -identify'
        cmd = shlex.split(cmd) + [video]
        output = Popen(cmd, stdout=PIPE, stderr=PIPE)
        return output.communicate()[0]

    def load(self, event=None):
        """Load a file to play in the GUI"""
        try:
            from tkFileDialog import askopenfilename
        except ImportError:
            # python 3
            from tkinter.filedialog import askopenfilename
        vid_name = askopenfilename()
        if vid_name:
            self.toggle_controls('normal', self.mp_ctrls)
            if self.pauseplay.get() == 'Pause':
                self.pauseplay.set('Play')
            self.run(vid_name)

    def set_container(self, video=None):
        """Get aspect ratio and set dimensions of video container.
           Called by run().
        """
        media_info = self.identify(video)
        asr = re.findall('ID_VIDEO_ASPECT=.*', str(media_info))
        # get last occurence as the first is 0.0 with mplayer
        if asr:
            asr = sorted(asr, reverse=True)[0].split('=')[1]
            try:
                asr = float(asr)
            except ValueError:
                asr = 0.0
        # get largest value as mplayer prints it out before playing file
        if asr and asr > 0.0:
            v_height = int(self.v_width/asr)
        else:
            # default to 4:3 if identify fails
            v_height = int(self.v_width/1.333)
        self.container.configure(width=self.v_width, height=v_height)

    def run(self, video):
        """Play video in this GUI using mplayer."""
        if video == None:
            self.toggle_controls('disabled', self.mp_ctrls)
            return
        elif not os.path.exists(video):
            try:
                from tkMessageBox import showerror
            except ImportError:
                from tkinter.messagebox import showerror
            showerror('Oops', video + ' does not exist')
            self.toggle_controls('disabled', self.mp_ctrls)
            self.master.master.withdraw()
        else:
            self.toggle_controls('enabled', self.mp_ctrls)
        # a new video has been loaded if no temp files, so make them
        if not os.path.exists(self.tempdir):
            self.make_tmps()
        # set the container size, then run the video
        self.set_container(video)
        command =  'mplayer -wid %s -nomouseinput -slave \
          -input nodefault-bindings:conf=/dev/null:file=%s \
          -edlout %s %s' \
          %(self.xid, self.cmd_pipe, self.editlist, self.args)
        self.command = shlex.split(command) + [video]

    def toggle_controls(self, state, widgets):
        """
        Enable/disable mplayer control widgets
        state is either 'normal' or 'disabled', widgets is an instance list
        """
        for widget in widgets:
            try:
                widget.configure(state=state)
            except tk.TclError:
                pass

    def poll(self):
        """
        Check mplayer log output for 'End of file' and restart the video.
        This is necessary because otherwise Tkinter has no way of knowing
        mplayer is still running.  In this way mplayer must be sent a 'quit'
        explicity before exit. Use 'on_eof()' to run custom command when
        'End of file' is found.
        """
        if not self.is_running.get():
            return
        f = open(self.log, 'r')
        # seek to 50 bytes from end of file
        try:
            f.seek(-50, 2)
            if '(End of file)' in f.read():
                self.on_eof()
                # check for is_running again as on_oef() might set it to false
                if self.is_running.get():
                    cmd = Popen(self.command, stderr=open(os.devnull, 'w'), \
                      stdout=open(self.log, "w"))
                    if self.show_osd:
                        self.send('osd 3')
                    self.pause()
        # if file does not contain 50 bytes, do nothing
        except IOError:
            pass
        self.master.after(100, self.poll)


    def on_eof(self):
        """
        Run when 'End of file' discovered in mplayer output by poll().
        Override to run custom commands.
        Note: change is_running() variable to false to prevent looping of video.
        """
        pass

    def send(self, text):
        """Send command to mplayer's slave fifo"""
        if self.is_running.get():
            getstatusoutput('echo "%s"  > %s' %(text, self.cmd_pipe))

    def pause(self):
        """Send pause to mplayer via slave and set button var to opposite value"""
        # mplayer's 'pause' pauses if playing and plays if paused
        # pauseplay ==Play in pause mode, and ==Pause in play mode (button text)
        if self.is_running.get():
            if self.pauseplay.get() == 'Pause':
                self.pauseplay.set('Play')
            else:
                self.pauseplay.set('Pause')
            self.send('pause')
        else:
            # start the video for the 1st time
            cmd = Popen(self.command, stderr=open(os.devnull, 'w'), stdout=open(self.log, "w"))
            self.is_running.set(True)
            self.poll()
            # show osd time and remaining time
            if self.show_osd:
                self.send('osd 3')
            self.pauseplay.set('Pause')

    def exit_mplayer(self): # called by [done] button
        """
        Close mplayer if it is running, then exit, running callback if it exists
        """
        # unpause so mplayer doesn't hang
        if self.is_running.get():
            if self.pauseplay.get() == 'Play':
                self.send('mute 1')
                self.send('pause')
            self.send('quit')
            self.is_running.set(False)
        time.sleep(0.3)
        self.confirm_exit()

    def confirm_msg(self):
        mess = "osd_show_text 'please exit mplayer first' 4000 3"
        if not self.show_osd:
            self.send('osd 3\n%s' %mess)
            self.after(2500, lambda:self.send('osd 0'))
        else:
            self.send(mess)

    def confirm_exit(self, event=None):
        """On exit, make sure that mplayer is not running before quit"""
        if self.is_running.get():
            self.confirm_msg()
        else:
            # run any custom commands on exit
            if callable(self.callback):
                self.callback()
            # remove temporary files and directory
            for f in self.cmd_pipe, self.log, self.editlist:
                if os.path.exists(f):
                    os.remove(f)
            if os.path.exists(self.tempdir):
                os.rmdir(self.tempdir)
            self.quit()

    def quit(self):
        """Quit the application.  This detroys the root window, exiting Tkinter"""
        self.destroy()


class SetChapters(VideoGui):
    """Elements for a GUI to set video chapter points using Mplayer"""
    def __init__(self, master, args='', title='', callback=None):
        """
        master
            Pack into this widget
        args
            Additional args to pass to mplayer
        title
            Window manager titlebar title (master must be root window for this)
        callback
            Function to run on application exit, run before temp file cleanup
        """
        VideoGui.__init__(self, master, args, title, callback)
        self.chapter_var = tk.StringVar()

    def add_widgets(self):
        """
        Add buttons to the Gui. root_frame has 'grab_set()' applied to it so
        make sure widgets go into this frame or they will not be functional!
        This function is called in draw()
        """
        # button frame and buttons
        button_frame = tk.Frame(self.root_frame)
        button_frame.pack(side='bottom', fill='x', expand=1)
        self.control_frame = tk.Frame(button_frame, borderwidth=1, relief='groove')
        self.control_frame.pack()
        exit_button = tk.Button(self.control_frame, command=self.exit_mplayer, text='done !')
        mark_button = tk.Button(self.control_frame, command=self.set_chapter,text='set chapter')
        pause_button = tk.Button(self.control_frame, command=self.pause,
                          width=12, textvariable=self.pauseplay)
        framestep_button = tk.Button(self.control_frame, text='step>',
                                                command=self.framestep)
        forward_button = tk.Button(self.control_frame, text='>>',
                                                command=self.forward)
        fast_forward_button = tk.Button(self.control_frame, text='>>>',
                                                command=self.fastforward)
        back_button = tk.Button(self.control_frame, text='<<', command=self.back)
        fast_back_button = tk.Button(self.control_frame, text='<<<', command=self.fast_back)
#        self.seek_scale = tk.Scale(seek_frame, from_=0, to=100, tickinterval=10,
#        orient='horizontal', label='Use slider to seek to point in file (%)')
#        self.seek_scale.bind('<ButtonRelease-1>', self.seek)
        # pack the buttons and scale in their frames
        mark_button.pack(side='bottom', fill='both', expand=1)
        #self.seek_scale.pack(side='left', fill='x', expand=1)
        exit_button.pack(side='left')
        fast_back_button.pack(side='left')
        back_button.pack(side='left')
        pause_button.pack(side='left', expand=1)
        framestep_button.pack(side='left')
        forward_button.pack(side='left')
        fast_forward_button.pack(side='left')
        self.mp_ctrls = self.control_frame.winfo_children()

#    def seek(self, event=None):
#        """Seek in video according to value set by slider"""
#        self.send('seek %s 3\n' %self.seek_scale.get())
#        self.after(500, lambda:self.seek_scale.set(0))

    def forward(self):
        """Seek forward 10 seconds and make sure button var is set to 'Pause'"""
        self.send('seek 10')
        self.pauseplay.set('Pause')

    def fastforward(self):
        """Seek forward 5 minutes and make sure button var is set to 'Pause'"""
        self.send('seek 300')
        self.pauseplay.set('Pause')

    def back(self):
        """Seek backward 10 seconds and make sure button var is set to 'Pause'"""
        self.send('seek -10')
        self.pauseplay.set('Pause')

    def fast_back(self):
        """Seek backward 10 seconds and make sure button var is set to 'Pause'"""
        self.send('seek -300')
        self.pauseplay.set('Pause')

    def framestep(self):
        """Step frame by frame forward and set button var to 'Play'"""
        self.send('pausing frame_step')
        self.pauseplay.set('Play')

    def set_chapter(self):
        """Send chapter mark (via slave) twice so mplayer writes the data.
           we only take the 1st mark on each line
        """
        for i in range(2):
            self.send('edl_mark')
        mess = "osd_show_text 'chapter point saved' 2000 3"
        if not self.show_osd:
            self.send('osd 3%s' %mess)
            self.after(2500, lambda:self.send('osd 0'))
        else:
            self.send(mess)


    def get_chapters(self):
        """Read mplayer's editlist to get chapter points and return
        HH:MM:SS.xxx format string, comma separated
        """
        # need a sleep to make sure mplayer gives up its data
        if not os.path.exists(self.editlist):
            return
        time.sleep(0.5)
        f = open(self.editlist)
        c = f.readlines()
        f.close()
        # if chapter_var has value, editlist has been reset.  Append value.
        # only 1st value on each line is taken (the 2nd makes mplayer write out)
        s = [ i.split()[0]  for i  in self.chapter_var.get().splitlines() if i]
        c.extend(s)
        times = [ float(shlex.split(i)[0]) for i in c ]
        chapters = ['00:00:00']
        for t in sorted(times):
            fraction = '.' + str(t).split('.')[1]
            chapters.append(time.strftime('%H:%M:%S', time.gmtime(t)) + fraction)
        if c:
            return '%s' %','.join(chapters)
    
    def on_eof(self):
        """
        Run when 'End of file' discovered in mplayer output by poll().
        """
        f = open(self.editlist)
        self.chapter_var.set(f.read())
        f.close()

class SetChaptersGui(SetChapters):
    """A standalone GUI to set video chapter points using SetChapters class"""
    def __init__(self, master, args='', title='', callback=None):
        """
        master
            Pack into this widget
        args
            Additional args to pass to mplayer
        title
            Window manager titlebar title (master must be root window for this)
        callback
            Function to run on application exit, run before temp file cleanup
        """
        SetChapters.__init__(self, master, args, title, callback)
        self.callback = self.print_chapters
        # bindings for exit
        self._root().protocol("WM_DELETE_WINDOW", self.exit)
        self._root().bind('<Control-q>', self.exit)


    def add_widgets(self):
        """Override add_widgets() from SetChapters, calling it first
        then adding widgets to it
        """
        SetChapters.add_widgets(self)
        self.result_frame = tk.Frame(self)
        self.text_frame = tk.Frame(self.result_frame)
        self.result_frame.pack(side='bottom', fill='x', expand=1)
        self.text_frame.pack(side='bottom', fill='x', expand=1)
        self.load_button = tk.Button(self.text_frame,
            command=self.load, text='Load Video')
        self.load_button.pack(side='left', padx=5)
        self.entry = tk.Entry(self.text_frame)
        self.entry.insert(0, "00:00:00.0")
        self.entry.pack(side='left', fill='x', expand=1)
        self.exit_button = tk.Button(self.text_frame,
            command=self.exit, text='Exit')
        self.exit_button.pack(side='left')
        # keep a list of mplayer controls so we can disable/enable them later
        self.mp_ctrls = self.control_frame.winfo_children()

    def load(self, event=None):
        """Load a file to play in the GUI"""
        try:
            from tkFileDialog import askopenfilename
        except ImportError:
            # python 3
            from tkinter.filedialog import askopenfilename
        vid_name = askopenfilename()
        if vid_name:
            self.toggle_controls('normal', self.mp_ctrls)
            # reset entry to 0 and chapter_var if we have a new video loaded
            self.entry.delete(0, tk.END)
            self.entry.insert(0, "00:00:00.0")
            if self.pauseplay.get() == 'Pause':
                self.pauseplay.set('Play')
            self.run(vid_name)

    def print_chapters(self):
        """
        Run get_chapters(), output result to stdout, entry box, and write
        to tempory file.  Disable mplayer controls.
        This functions as the callback on mplayer exit.
        """
        if self.get_chapters():
            try:
                from tkMessageBox import askyesno, showinfo
            except ImportError:
                # python 3
                from tkinter.messagebox import askyesno, showinfo
            output = self.get_chapters()
            stdout.write(output + '\n')
            self.entry.delete(0, tk.END)
            self.entry.insert(0, output)
            self.toggle_controls('disabled', self.mp_ctrls)
            string1 = "Chapter string will be saved to: "
            string2 = '\nDo you wish to save it ?'
            #string2 += '  Choose "Yes" to save it.'
            # get basename of video, remove extension, add a '_'
            vid = os.path.basename(self.command[-1]) + '_'
            tmpfile = mkstemp(prefix=vid)
            if askyesno(title="Save chapter string",
              message= '%s%s%s' %(string1, tmpfile[1], string2)):
                save_msg = "%s%s\n" %('Saved ', tmpfile[1])
                showinfo(title='Chapters saved', message=save_msg)
                if os.path.exists(tmpfile[1]):
                    f = open(tmpfile[1], 'w')
                    f.write(output)
                    f.close()
            else:
                    os.remove(tmpfile[1])

    def quit(self):
        """Override quit() from base class, as standalone uses exit()"""
        # disable controls, because tempdir has been deleted
        self.toggle_controls('disabled', self.mp_ctrls)
        if self.chapter_var.get():
            self.chapter_var.set('')

    def exit(self, event=None):
        """Exit the GUI after confirming that mplayer is not running."""
        if self.is_running.get():
            self.confirm_msg()
        else:
            self.master.destroy()


# class for control that allow setting chapter points
class Chapters(ListToOne):
    """A popup version of the ListToOne Control, that also
    allows setting chapter points with a mplayer GUI
    (SetChapters).  This Control is specific to the tovid GUI.
    """
    def __init__(self,
                 parent,
                 label="ListToOne",
                 option='',
                 default=None,
                 help='',
                 filter=lambda x: x,
                 side='left',
                 control=Text(),
                 text='',
                 **kwargs):
        """initialize Chapters
        text
           The text for the button that calls mplayer
        For other options see help(control.ListToOne) 
        """
        ListToOne.__init__(self, parent, label, option, default, help,
                          control, filter, side, **kwargs)
        self.text = text
        self.parent = parent
        self.top_width = 540
        self.top_height = 540

    def draw(self, master):
        """Initialize Toplevel popup, video/chapters lists, and mplayer GUI.
        Only pack the video/chapter lists (_SubList).  Withdraw until called.
        """
        chapters_button = tk.Button(master, text='edit', command=self.popup)
        chapters_button.pack()
        # popup to hold lists and mplayer
        self.top = tk.Toplevel(master)
        self.top.withdraw()
        self.top.minsize(540, 540)
        self.top.title('Chapters')
        # text and label for instructions
        txt = 'Auto chapters:\n' + \
        '   1. Enter single value (integer) for all videos on 1st line.\n' + \
        '   2. Or, use a different auto chapter value for each video.\n' + \
        'HH:MM:SS format:\n' + \
        '   1. Enter timecode for each video, eg. 00:00:00,00:05:00 ...\n' + \
        '   2. Or use the "set with mplayer" button to use a GUI.'
        self.top.label = tk.Label(self.top, text=txt, justify=tk.LEFT)
        self.top.label.pack(side=tk.TOP)
        self.top_button = tk.Button(self.top, text='Okay', command=self.top.withdraw)
        # frames to hold the _Sublist and mplayer
        self.sublist_frame = tk.Frame(self.top)
        self.mplayer_frame = tk.Frame(self.top)
        # bindings for initial toplevel window
        self.top.protocol("WM_DELETE_WINDOW", self.top.withdraw)
        self.top.bind('<Escape>', self.withdraw_popup)
        # pack the _SubList frame and draw _SubList
        self.sublist_frame.pack(expand=1, fill=tk.BOTH)
        _SubList.draw(self, self.sublist_frame, allow_add_remove=False)
        self.top_button.pack(side='bottom')

        # a button to allow setting chapters in mplayer GUI
        button = tk.Button(self.control, text=self.text,
        command=self.run_mplayer, state='disabled')
        button.pack(side='left')
        # 1:1, parent listbox is linked to this one
        self.parent_listbox.link(self.listbox)
        # Add callbacks to handle changes in parent
        self.add_callbacks()

    def get_geo(self, widget):
        """Get geometry of a widget.
        Returns List (integers): width, height, Xpos, Ypos
        """
        geo = widget.winfo_geometry()
        return  [ int(x) for x in geo.replace('x', '+').split('+') ]

    def center_popup(self, master, width, height):
        """Get centered screen location of popup, relative to master.
        Returns geometry string in form of 'WxH+Xpos+Ypos'
        """
        root_width, root_height, rootx, rooty = self.get_geo(master)
        xoffset = (root_width - width) / 2
        # subtract 15 pixels for WM titlebar (still looks OK if there is none)
        yoffset = ((root_height - height) / 2 ) - 15
        return '%dx%d+%d+%d' %(width, height, rootx + xoffset, rooty + yoffset)
        
    def popup(self):
            """Popup the list of chapters, with button to run mplayer GUI"""
            videolist = self.parent_listbox
            self.top.transient(self.parent._root())
            w = self.top_width
            h = self.top_height
            self.top.geometry(self.center_popup(self.parent._root(), w, h))
            self.after(100, lambda:self.top.deiconify())
            if videolist.items.count() and not videolist.selected.get():
                self.parent_listbox.select_index(0)
    
    def run_mplayer(self, event=None):
        """Run the mplayer GUI to set chapters"""
        selected = self.parent_listbox.selected.get()
        if selected:
            # initialize mplayer GUI
            self.mpl = SetChapters(self.mplayer_frame, '-osdlevel 3', '', self.on_exit)
            self.mpl.pack()
            # unpack label
            self.sublist_frame.pack_forget()
            self.top.label.pack_forget()
            self.top_button.pack_forget()
            self.mplayer_frame.pack()
            self.mpl.run(selected)
            # disable close button while mplayer running, both top and _root()
            self.top.protocol("WM_DELETE_WINDOW", self.mpl.confirm_exit)
            self.master._root().protocol("WM_DELETE_WINDOW", self.mpl.confirm_exit)
            # disable usage of _root() window while mplayer running
            self.top.grab_set()
            self.top.bind('<Escape>', self.mpl.confirm_exit)

    def on_exit(self):
        """Callback run when mplayer GUI exits. This sets the chapters list
        to the timecodes set by the mplayer gui, repacks the label and the
        list frame, sets/resets bindings, and releases the grab_set().
        """
        if self.mpl.get_chapters():
            self.control.variable.set(self.mpl.get_chapters())
        self.mplayer_frame.pack_forget()
        # repack label
        self.top.label.pack(side=tk.TOP)
        self.sublist_frame.pack(expand=1, fill=tk.BOTH)
        self.top_button.pack(side='bottom')
        self.top.protocol("WM_DELETE_WINDOW", self.top.withdraw)
        self.top.grab_release()
        self.master._root().protocol("WM_DELETE_WINDOW", self.master._root().confirm_exit)
        self.top.bind('<Escape>', self.withdraw_popup)

    def withdraw_popup(self, event=None):
        """Withdraw the chapters popup window"""
        self.top.withdraw()

    def select(self, index, value):
        """Select an item in the list and enable editing.
        List select method overriden in order to clear entry selection.
        """
        List.select(self, index, value)
        # don't make it easy to erase a chapters string with a keypress
        self.control.selection_clear()

# Define a few supporting functions
def to_title(filename):
    basename = os.path.basename(filename)
    firstdot = basename.find('.')
    if firstdot >= 0:
        return basename[0:firstdot]
    else:
        return basename

def strip_all(filename):
    return ''

def find_masks(dir, pattern):
    file_list=[]
    ext = pattern.replace('*', '')
    for path, dirs, files in os.walk(os.path.abspath(dir)):
        for filename in fnmatch.filter(files, pattern):
            file_list+=[ filename.replace(ext, '') ]
    return file_list

def nodupes(seq):
    noDupes = []
    [noDupes.append(i) for i in seq if not noDupes.count(i)]
    return noDupes

def get_loadable_opts(opts):
    good_opts = []
    bad_opts = []
    add_opt = False
    for index, opt in enumerate(opts):
        if opt.startswith('-'):
            if opt in no_load_opts:
                add_opt = False
            else:
                add_opt = True
                # -colour is not loadable, but is a valid todisc option
                if '-colour' in opt:
                    opt = opt.replace('-colour','-color')
        if add_opt:
            good_opts.append(opt)
        else:
            bad_opts.append(opt)

    return [good_opts, bad_opts]

# this is part of the HACK in working around broken load_args
# load_script was lifted from gui.py in metagui, with mods
def load_script(filename):
    #"""Load current script options and arguments from a text file.
    #"""
    # Read lines from the file and reassemble the command
    command = ''
    for line in open(filename, 'r'):
        line = line.strip()
        # Discard comment lines and PATH assignment
        if line and not (line.startswith('#') or line.startswith('PATH=')):
            command += line.rstrip('\\')
    # Split the full command into arguments, according to shell syntax
    args = shlex.split(command)

    # Ensure the expected program is being run
    program = args.pop(0)
    if program != 'todisc':
        raise ValueError("This script runs '%s', expected '%s'" %
              (program, 'todisc'))
    return args


def filter_args(master=None, args=None):
    # HACK
    # we need to sanitize the GUI args as load_args in metagui is broken
    # this function and imports of it can be removed when it is fixed
    a = get_loadable_opts(args)
    args = a[0]
    if a[1]:
        #import Tkinter as tk
        from textwrap import dedent
        heading = '''
        The tovid GUI did not load some of your options.
        This is normal as loading options externally
        is experimental. Sorry for the inconvenience.

        You can copy this for reference as you will need
        to set the following options yourself in the GUI:

        '''

        #u = unloadable_opts[:]
        #data = ['\n' + x + '\n' if x.startswith('-') else x for x in u]
        nonloadable = []
        for i in a[1]:
            if i.startswith('-'):
                nonloadable.append('\n' + i)
            elif ' ' in i:
                nonloadable.append("'" + i + "'")
            else:
                nonloadable.append(i)
        #data = ' '.join(nonloadable)
        master.title('Important!')
        info_msg = CopyableInfo(master, 'Important', dedent(heading),
                                     ' '.join(nonloadable), tovid_icon)
        info_msg.mainloop()
    return args


from libtovid.metagui.support import show_icons
from libtovid.metagui import Style
class CopyableInfo(tk.Frame):
    def __init__(self, master, title='', heading='', data='', icon=None):
        tk.Frame.__init__(self, master)
        self.pack()
        master.minsize(300, 300)
        text_frame = tk.Frame(self)
        text_frame.pack(side='top')
        style = Style()
        inifile = os.path.expanduser('~/.metagui/config')
        if os.path.exists(inifile):
            style.load(inifile)
            font = style.font or None
        else:
            font = None
        text_widget = tk.Text(text_frame, width=50, height=20, wrap='word')
        text_widget["bg"] = self.cget('background')
        text_widget["relief"] = 'flat'
        master["relief"] = 'flat'
        if font:
            text_widget["font"] = font
        text_widget.pack(side='top', fill='both', expand=1)
        text_widget.insert('1.0',heading)
        #text_widget.bind("<1>", self.set_focus(self, text_widget))
        text_widget.insert('end',data)
        text_widget['state'] = 'disabled'
        lines = int(text_widget.index('end-1c').split('.')[0]) 
        text_widget['height'] = lines + 3
        rclick = RightClickMenu(text_widget)
        text_widget.bind("<3>", rclick)
        exit_button = tk.Button(self, text='Close',
                       command=lambda:self._quit())
        #exit_button.focus_set()
        exit_button.pack(side='bottom')
        master.bind('<Control-q>', lambda event=None: self._quit())
        master.bind('<Escape>', lambda event=None: self._quit())
        if icon:
            show_icons(self.master, icon)
        master.update_idletasks() # update geometry
        # center the popup
        m = master
        x = int((m.winfo_screenwidth() / 2) - (m.winfo_width() /2))
        y = int((m.winfo_screenheight() / 2) - int((m.winfo_height() /2)))
        master.geometry('+%s+%s' %(x, y))



    def _quit(self, event=None):
        self.master.destroy()
        #self.master.destroy()


class RightClickMenu(object):
    # thanks to 'paperrobot' at:
    #  http://paperrobot.wordpress.com for this gem
    """
    Simple widget to add basic right click menus to entry widgets.

    usage:

    rclickmenu = RightClickMenu(some_entry_widget)
    some_entry_widget.bind("<3>", rclickmenu)

    Replace all Tix references with Tkinter and this will still work fine.
    """
    def __init__(self, parent):
        self.parent = parent
        # bind Control-A to select_all() to the widget.  Others work ok without
        self.parent.bind("<Control-a>", lambda e: self.select_all(), add='+')
        self.parent.bind("<Control-A>", lambda e: self.select_all(), add='+')
    def __call__(self, event):
        # grab focus of the entry widget.  this way you can see
        # the cursor and any marking selections
        self.parent.focus_force()
        self.build_menu(event)
    def build_menu(self, event):
        self.menu = tk.Menu(self.parent, tearoff=0)
        #self.parent.bind("<Button-1>", self.close_menu)
        #self.parent.bind("<Escape>", self.close_menu)
        if self.parent.tag_ranges("sel"):
            self.menu.add_command(
                label="Copy",
                command=lambda: self.parent.event_generate("<<Copy>>"))
        else:
            self.menu.add_command(label="Copy", state=tk.DISABLED)
        self.menu.add_command(label="Cut", state=tk.DISABLED)
        self.menu.add_command(label="Paste", state=tk.DISABLED)
        # make things pretty with a horizontal separator
        self.menu.add_separator()
        self.menu.add_command(label="Select All", command=self.select_all)
        self.menu.tk_popup(event.x_root, event.y_root)
    def select_all(self):
        self.parent.tag_add(tk.SEL, 1.0, tk.END)
        # return 'break' because doesn't work otherwise.
        return 'break'



# List of file-type selections for Filename controls
image_filetypes = [filetypes.image_files]
image_filetypes.append(filetypes.all_files)
#image_filetypes.extend(filetypes.match_types('image'))  # confusing
# video file-types from filetypes needs some additions
v_filetypes = 'm2v vob ts '
v_filetypes += filetypes.get_extensions('video').replace('*.', '')
v_filetypes += ' mp4 mpeg4 mp4v divx mkv ogv ogm ram rm rmvb wmv webm'
vid_filetypes = filetypes.new_filetype('Video files', v_filetypes)
video_filetypes = [vid_filetypes]
video_filetypes += [filetypes.all_files]

# some selectors can use video or audio
av_filetypes = [ filetypes.all_files, filetypes.audio_files, vid_filetypes ]

# some selectors can use image or video
visual_filetypes = [ filetypes.all_files, filetypes.image_files, vid_filetypes ]

# DVD video
dvdext = 'vob mpg mpeg mpeg2'
dvd_video_files = [ filetypes.new_filetype('DVD video files', dvdext) ]

# Users can use their own thumb masks.  Add to thumb mask control drop-down
masks = [ 'none', 'normal', 'oval', 'vignette', 'plectrum', 'arch', 'spiral', \
'blob', 'star', 'flare' ]

# some options can not be loaded by the gui:
# options with multiple choices for args, for example: none, single, multiple
# this is a list of 'unloadable' options for the GUI
no_load_opts = ['-chapter-titles', '-menu-fade', '-loop', '-video-pause', '-group-video-pause', '-slide-blur', '-slide-border', '-slide-frame', '-chain-videos', '-showcase', '-titles-font-deco', '-chapters', '-subtitle-lang', '-audio-channel', '-audio-lang', '-seek', '-showcase-seek', '-bg-video-seek', '-bgvideo-seek', '-bg-audio-seek', '-bgaudio-seek', '-submenu-audio-seek', '-rotate-thumbs', '-tile3x1', '-tile-3x1', '-tile4x1', '-tile-4x1']

# $PREFIX/lib/tovid is already added to end of PATH
os_path = os.environ['PATH'].rsplit(':')
sys_dir = os_path[-1] + '/masks'
home_dir = os.path.expanduser("~") + '/.tovid/masks'
for dir in sys_dir, home_dir:
    masks.extend(find_masks(dir, '*.png'))
thumb_masks =  '|'.join(nodupes(masks))
tovid_prefix = getoutput('tovid -prefix')
tovid_icon = os.path.join(tovid_prefix, 'lib', 'tovid', 'tovid.png')

# default heading text for tovid GUI (opening pane)
heading_text = 'You can author (and burn) a DVD with a simple menu '
heading_text += 'using ONLY this "Basic" pane\n'
heading_text += 'Required: video files (and/or slideshows) and "Output name"'

### configuration for titleset wizard ( 'tovid titlesets' )
wizard = os.getenv('METAGUI_WIZARD') or '0'
if wizard == '1':
    # General options
    heading_text = 'General options applying to all titlesets.\n'
    heading_text += 'Required: "Output name" at bottom of window'
elif wizard == '2':
    # Root menu options
    heading_text = 'Root menu options.  Use any options that do not depend\n'
    heading_text += 'on video files or slideshows being loaded. Required: None'
    heading_text += '\nChange "Menu title" from the default to suit your needs'
elif wizard >= '3':
    # titleset options
    heading_text = 'Options for titleset %s\n' %(int(wizard) - 2)
    heading_text += 'Required: 1 or more videos and/or slideshows'

if wizard == '2':
    # Root menu options
    _files = List('Do not load any files for Root menu !', '-files', None,
        '',
        Filename('', filetypes=video_filetypes))
    _files_and_titles = ListToOne(_files, 'Root menu link titleset titles',
        '-titles', None,
        'The titles for the Root menu link titleset titles',
        Text(),
        filter=to_title)
else:
    _files = List('Video files', '-files', None,
        'List of video files to include on the disc',
        Filename('', filetypes=video_filetypes))
    _files_and_titles = ListToOne(_files, 'Video titles', '-titles', None,
        'Titles to display in the main menu, one for each video file on the disc',
        Text(),
        filter=to_title)

if int(wizard) and wizard != '1':
    # 'Output name' only for General Options ( '1' )
    _out = Label(' ')
else:
    _out = Filename('Output name', '-out', '',
        'Name to use for the output directory where the disc will be created.',
        'save', 'Choose an output name')



"""Provides a Python interface to the functionality of tovid.
"""

__all__ = [
    # Subdirectories
    'guis',
    'metagui',
    # .py files
    'cli',
    'odict',
    'utils',
]

import os
from sys import version_info

# Python < 3.x
try:
    from ConfigParser import ConfigParser
# Python 3.x
except ImportError:
    from configparser import ConfigParser

# Python 3.x compatibility assignments
if version_info[0] < 3: # python 3.x
    unicode = unicode
    basestring = basestring
    xrange = xrange
else: # Python 3.x
    unicode = str
    basestring = str
    xrange = range

# Configuration file reader/writer
class Config (ConfigParser):
    """Interface for reading/writing tovid configuration files. Just a wrapper
    around the standard library ConfigParser. Example usage::

        config = libtovid.Config()
        config.get('DEFAULT', 'work_dir')
        config.set('tovid', 'method', 'ffmpeg')
        config.save()

    See the ConfigParser documentation for details.
    """
    # Dictionary of suite-wide configuration defaults
    DEFAULTS = {
        'work_dir': '/tmp',
        'output_dir': '/tmp'}

    def __init__(self):
        """Load configuration from ~/.tovid/config."""
        ConfigParser.__init__(self, self.DEFAULTS)
        self.filename = os.path.expanduser('~/.tovid/config')
        self.read(self.filename)

    def save(self):
        """Save the configuration to the current filename."""
        outfile = open(self.filename, 'w')
        outfile.write('# tovid configuration file\n\n')
        self.write(outfile)
        outfile.close()



# Logging class
class Log:
    """Logging class, with five severity levels.
    """
    # Increasing levels of severity
    _levels = {
        'debug': 1,
        'info': 2,
        'warning': 3,
        'error': 4,
        'critical': 5}

    def __init__(self, level='info'):
        """Create a logger with the given severity level."""
        self.level = level

    def message(self, level, text):
        """Print a message if it's at the current severity level or higher."""
        if Log._levels[level] >= Log._levels[self.level]:
            print(text)

    def debug(self, text):
        """Log a debugging message."""
        self.message('debug', text)

    def info(self, text):
        """Log an informational message."""
        self.message('info', text)

    def warning(self, text):
        """Log a warning message."""
        self.message('warning', "WARNING: " + text)

    def error(self, text):
        """Log an error message."""
        self.message('error', "ERROR: " + text)

    def critical(self, text):
        """Log a critical error message."""
        self.message('critical', "CRITICAL: " + text)



# Global logger
log = Log()



# makemenu.py

from libtovid.metagui import *

# Menu TV system
_ntsc = Flag('NTSC', '-ntsc', True)
_ntscfilm = Flag('NTSC Film', '-ntscfilm', False)
_pal = Flag('PAL', '-pal', False)
TVSYS = FlagGroup('TV system', 'exclusive', _ntsc, _ntscfilm, _pal)

# Menu format
_vcd = Flag('VCD', '-vcd', False)
_svcd = Flag('SVCD', '-svcd', False)
_dvd = Flag('DVD', '-dvd', True,
    'Create a menu in DVD format. Extra features are enabled for DVD menus.',
    enables=['-button', '-button-font', '-button-outline',
        '-highlightcolor', '-selectcolor'])

FORMAT = FlagGroup('Menu format', 'exclusive', _vcd, _svcd, _dvd)

# DVD-only options (enabled by -dvd)
_button = Text('Button character', '-button', '>',
    'A single character to use for a menu cursor, or one of '
    '"play", "movie", or "utf8 xxxx".')

_highlightcolor = Color('Highlight color', '-highlightcolor')
_selectcolor = Color('Select color', '-selectcolor')
_button_outline = Color('Outline color', '-button-outline')
_button_font = Font('Button font', '-button-font')

DVD_ONLY = VPanel('DVD menu elements',
    _button,
    _button_font,
    _button_outline,
    _highlightcolor,
    _selectcolor,
)

# Background
_background = Filename('BG image', '-background')
_crop = Flag('Crop', '-crop', True)
_scale = Flag('Scale', '-scale', False)
_audio = Filename('BG audio', '-audio')
_length = Number('Duration', '-length', 30,
    'Menu duration', 1, 600, 'seconds')
_nosafearea = Flag('No safe area', '-nosafearea', False)

# Fonts
_menu_title = Text('Menu title', '-menu-title')
_font = Font('Menu font', '-font')
_fontsize = Number('Font size', '-fontsize', 24,
    'Menu font size', 10, 100, 'pixels')
_menu_title_fontsize = Number('Title font size', '-menu-title-fontsze', 32,
    'Menu title (heading) font size', 10, 100, 'pixels')
_fontdeco = Text('Font decoration', '-fontdeco')
_align = Choice('Text alignment', '-align', 'left',
    'Align/justify the menu text', 'left|center|middle|right')
_textcolor = Color('Menu text color', '-textcolor')

# Behavior
_debug = Flag('Debug', '-debug', False)
_overwrite = Flag('Overwrite', '-overwrite', False)
_noask = Flag('No prompting', '-noask', False)
_quiet = Flag('Quiet', '-quiet', False)



TITLES = List("Titles", '<titles>', '',
    "Video titles to include on the menu",
    Filename())

OUT_PREFIX = Filename("Output name", '-out', '',
    "Name of output file (will be given a .mpg extension)",
    'save', "Choose an output name")


MAIN = VPanel('Makemenu options',
    HPanel('', FORMAT, DVD_ONLY, TVSYS),
    VPanel('Background',
        _background,
        _audio,
        HPanel('',
            FlagGroup('Fit image', 'exclusive', _crop, _scale),
            _nosafearea,
            _length
        ),
    ),

    _menu_title,
    _menu_title_fontsize,

    _textcolor,
    _font,
    _fontsize,
    _align,
    _fontdeco,

    _debug,
    _overwrite,
    _noask,
    _quiet,

    TITLES,
    OUT_PREFIX,
)

def run():
    app = Application('makemenu', MAIN)
    gui = GUI("tovid metagui", 640, 720, app)
    gui.run()

if __name__ == '__main__':
    run()




"""Contains Tkinter Variable subclasses for List and Dict variables.

This module exists to supplement the built-in Tkinter Variable types,
which do not provide `list` and `dict` equivalents.
"""

__all__ = [
    'ListVar',
    'DictVar',
    'VAR_TYPES',
]

# Python < 3.x
try:
    import Tkinter as tk
# Python 3.x
except ImportError:
    import tkinter as tk

from libtovid.odict import Odict

class ListVar (tk.Variable):
    """A tk Variable suitable for associating with Listboxes.
    """
    def __init__(self, master=None, items=None):
        """Create a ListVar with a given master and initial list of items.
        """
        tk.Variable.__init__(self, master)
        if items:
            self.set(items)
        self.callbacks = {'insert': [], 'remove': [], 'select': []}


    def get(self):
        """Get the entire list of values.
        """
        return list(tk.Variable.get(self))


    def set(self, new_list):
        """Set the entire list of values.
        """
        tk.Variable.set(self, tuple(new_list))


    def __getitem__(self, index):
        """Get a list value using list-index syntax: ``listvar[index]``.
        """
        return self.get()[index]


    def __setitem__(self, index, value):
        """Set a list value using list-index syntax: ``listvar[index] = value``.
        """
        current = self.get()
        current[index] = value
        self.set(current)


    def insert(self, index, item):
        """Insert an item into the list at the specified index.
        """
        items = self.get()
        items.insert(index, item)
        self.set(items)


    def remove(self, item):
        """Remove the item from the list, if it exists.
        """
        items = self.get()
        items.remove(item)
        self.set(items)


    def append(self, item):
        """Append an item to the list.
        """
        self.insert(-1, item)


    def pop(self, index=-1):
        """Pop an item off the list and return it.
        """
        items = self.get()
        item = items.pop(index)
        self.set(items)
        return item


    def index(self, value):
        """Return the index of the given value.
        """
        items = self.get()
        return items.index(value)


    def count(self):
        """Return the number of items in the list.
        """
        return len(self.get())


class DictVar (tk.Variable):
    """A tk Variable for storing a dictionary of values.
    """
    def __init__(self, master=None, keys=None, values=None):
        """Create a DictVar with a given master and initial keys/values.
        """
        tk.Variable.__init__(self, master)
        self.set(Odict(keys, values))


    def __getitem__(self, key):
        """Get a dict value using keyword syntax: ``dictvar[key]``.
        """
        current = self.get()
        print("DictVar: %s" % current)
        return current[key]


    def __setitem__(self, key, value):
        """Set a dict value using keyword syntax: ``dictvar[key] = value``.
        """
        current = self.get()
        current[key] = value
        self.set(current)


    def pop(self, key):
        """Pop the item at the given key, and return it.
        """
        current = self.get()
        value = current.pop(key)
        self.set(current)
        return value


    def get(self):
        """Return the entire dictionary of keys/values as an
        `~libtovid.odict.Odict`.
        """
        # Convert from tuple
        tup = tk.Variable.get(self)
        if tup:
            keys, values = zip(*tup)
        else:
            keys, values = [], []
        return Odict(keys, values)


    def set(self, new_dict):
        """Set the entire dictionary of keys/values. If ``new_dict`` is empty, or
        is not a `dict` or `~libtovid.odict.Odict`, an empty `dict` is used.
        """
        if not isinstance(new_dict, dict) and not isinstance(new_dict, Odict):
            new_dict = {}
        if not new_dict:
            new_dict = {}
        # Convert to a tuple of (key, value) pairs
        tup = tuple([(key, value) for key, value in new_dict.items()])
        tk.Variable.set(self, tup)


# Map python types to Tkinter variable types
VAR_TYPES = {
    str: tk.StringVar,
    bool: tk.BooleanVar,
    int: tk.IntVar,
    float: tk.DoubleVar,
    list: ListVar,
    dict: DictVar,
}
